import torch
import pytest

from pina import LabelTensor, Condition
from pina.condition import (
    TensorInputGraphTargetCondition,
    TensorInputTensorTargetCondition,
    GraphInputGraphTargetCondition,
    GraphInputTensorTargetCondition,
)
from pina.condition import (
    InputTensorEquationCondition,
    InputGraphEquationCondition,
    DomainEquationCondition,
)
from pina.condition import (
    TensorDataCondition,
    GraphDataCondition,
)
from pina.domain import CartesianDomain
from pina.equation.equation_factory import FixedValue
from pina.graph import RadiusGraph

example_domain = CartesianDomain({"x": [0, 1], "y": [0, 1]})

input_tensor = torch.rand((10, 3))
target_tensor = torch.rand((10, 2))
input_lt = LabelTensor(torch.rand((10, 3)), ["x", "y", "z"])
target_lt = LabelTensor(torch.rand((10, 2)), ["a", "b"])

x = torch.rand(10, 20, 2)
pos = torch.rand(10, 20, 2)
radius = 0.1
input_graph = [
    RadiusGraph(
        x=x_,
        pos=pos_,
        radius=radius,
    )
    for x_, pos_ in zip(x, pos)
]
target_graph = [
    RadiusGraph(
        x=x_,
        pos=pos_,
        radius=radius,
    )
    for x_, pos_ in zip(x, pos)
]

x = LabelTensor(torch.rand(10, 20, 2), ["u", "v"])
pos = LabelTensor(torch.rand(10, 20, 2), ["x", "y"])
radius = 0.1
input_graph_lt = [
    RadiusGraph(
        x=x[i],
        pos=pos[i],
        radius=radius,
    )
    for i in range(len(x))
]
target_graph_lt = [
    RadiusGraph(
        x=x[i],
        pos=pos[i],
        radius=radius,
    )
    for i in range(len(x))
]

input_single_graph = input_graph[0]
target_single_graph = target_graph[0]


def test_init_input_target():
    cond = Condition(input=input_tensor, target=target_tensor)
    assert isinstance(cond, TensorInputTensorTargetCondition)
    cond = Condition(input=input_tensor, target=target_tensor)
    assert isinstance(cond, TensorInputTensorTargetCondition)
    cond = Condition(input=input_tensor, target=target_graph)
    assert isinstance(cond, TensorInputGraphTargetCondition)
    cond = Condition(input=input_graph, target=target_tensor)
    assert isinstance(cond, GraphInputTensorTargetCondition)
    cond = Condition(input=input_graph, target=target_graph)
    assert isinstance(cond, GraphInputGraphTargetCondition)

    cond = Condition(input=input_lt, target=input_single_graph)
    assert isinstance(cond, TensorInputGraphTargetCondition)
    cond = Condition(input=input_single_graph, target=target_lt)
    assert isinstance(cond, GraphInputTensorTargetCondition)
    cond = Condition(input=input_graph, target=target_graph)
    assert isinstance(cond, GraphInputGraphTargetCondition)
    cond = Condition(input=input_single_graph, target=target_single_graph)
    assert isinstance(cond, GraphInputGraphTargetCondition)

    with pytest.raises(ValueError):
        Condition(input_tensor, input_tensor)
    with pytest.raises(ValueError):
        Condition(input=3.0, target="example")
    with pytest.raises(ValueError):
        Condition(input=example_domain, target=example_domain)

    # Test wrong graph condition initialisation
    input = [input_graph[0], input_graph_lt[0]]
    target = [target_graph[0], target_graph_lt[0]]
    with pytest.raises(ValueError):
        Condition(input=input, target=target)

    input_graph_lt[0].x.labels = ["a", "b"]
    with pytest.raises(ValueError):
        Condition(input=input_graph_lt, target=target_graph_lt)
    input_graph_lt[0].x.labels = ["u", "v"]


def test_init_domain_equation():
    cond = Condition(domain=example_domain, equation=FixedValue(0.0))
    assert isinstance(cond, DomainEquationCondition)
    with pytest.raises(ValueError):
        Condition(example_domain, FixedValue(0.0))
    with pytest.raises(ValueError):
        Condition(domain=3.0, equation="example")
    with pytest.raises(ValueError):
        Condition(domain=input_tensor, equation=input_graph)


def test_init_input_equation():
    cond = Condition(input=input_lt, equation=FixedValue(0.0))
    assert isinstance(cond, InputTensorEquationCondition)
    cond = Condition(input=input_graph_lt, equation=FixedValue(0.0))
    assert isinstance(cond, InputGraphEquationCondition)
    with pytest.raises(ValueError):
        cond = Condition(input=input_tensor, equation=FixedValue(0.0))
    with pytest.raises(ValueError):
        Condition(example_domain, FixedValue(0.0))
    with pytest.raises(ValueError):
        Condition(input=3.0, equation="example")
    with pytest.raises(ValueError):
        Condition(input=example_domain, equation=input_graph)


test_init_input_equation()


def test_init_data_condition():
    cond = Condition(input=input_lt)
    assert isinstance(cond, TensorDataCondition)
    cond = Condition(input=input_tensor)
    assert isinstance(cond, TensorDataCondition)
    cond = Condition(input=input_tensor, conditional_variables=torch.tensor(1))
    assert isinstance(cond, TensorDataCondition)
    cond = Condition(input=input_graph)
    assert isinstance(cond, GraphDataCondition)
    cond = Condition(input=input_graph, conditional_variables=torch.tensor(1))
    assert isinstance(cond, GraphDataCondition)



"""Module for the Supervised solver interface."""

from abc import abstractmethod

import torch

from torch.nn.modules.loss import _Loss
from ..solver import SolverInterface
from ...utils import check_consistency
from ...loss.loss_interface import LossInterface
from ...condition import InputTargetCondition


class SupervisedSolverInterface(SolverInterface):
    r"""
    Base class for Supervised solvers. This class implements a Supervised Solver
    , using a user specified ``model`` to solve a specific ``problem``.

    The ``SupervisedSolverInterface`` class can be used to define
    Supervised solvers that work with one or multiple optimizers and/or models.
    By default, it is compatible with problems defined by
    :class:`~pina.problem.abstract_problem.AbstractProblem`,
    and users can choose the problem type the solver is meant to address.
    """

    accepted_conditions_types = InputTargetCondition

    def __init__(self, loss=None, **kwargs):
        """
        Initialization of the :class:`SupervisedSolver` class.

        :param AbstractProblem problem: The problem to be solved.
        :param torch.nn.Module loss: The loss function to be minimized.
            If ``None``, the :class:`torch.nn.MSELoss` loss is used.
            Default is `None`.
        :param kwargs: Additional keyword arguments to be passed to the
            :class:`~pina.solver.solver.SolverInterface` class.
        """
        if loss is None:
            loss = torch.nn.MSELoss()

        super().__init__(**kwargs)

        # check consistency
        check_consistency(loss, (LossInterface, _Loss), subclass=False)

        # assign variables
        self._loss_fn = loss

    def optimization_cycle(self, batch):
        """
        The optimization cycle for the solvers.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :return: The losses computed for all conditions in the batch, casted
            to a subclass of :class:`torch.Tensor`. It should return a dict
            containing the condition name and the associated scalar loss.
        :rtype: dict
        """
        condition_loss = {}
        for condition_name, points in batch:
            condition_loss[condition_name] = self.loss_data(
                input=points["input"], target=points["target"]
            )
        return condition_loss

    @abstractmethod
    def loss_data(self, input, target):
        """
        Compute the data loss for the Supervised. This method is abstract and
        should be override by derived classes.

        :param input: The input to the neural network.
        :type input: LabelTensor | torch.Tensor | Graph | Data
        :param target: The target to compare with the network's output.
        :type target: LabelTensor | torch.Tensor | Graph | Data
        :return: The supervised loss, averaged over the number of observations.
        :rtype: LabelTensor | torch.Tensor | Graph | Data
        """

    @property
    def loss(self):
        """
        The loss function to be minimized.

        :return: The loss function to be minimized.
        :rtype: torch.nn.Module
        """
        return self._loss_fn



"""Module for the solver classes."""

__all__ = [
    "SolverInterface",
    "SingleSolverInterface",
    "MultiSolverInterface",
    "PINNInterface",
    "PINN",
    "GradientPINN",
    "CausalPINN",
    "CompetitivePINN",
    "SelfAdaptivePINN",
    "RBAPINN",
    "SupervisedSolverInterface",
    "SupervisedSolver",
    "ReducedOrderModelSolver",
    "DeepEnsembleSolverInterface",
    "DeepEnsembleSupervisedSolver",
    "DeepEnsemblePINN",
    "GAROM",
]

from .solver import SolverInterface, SingleSolverInterface, MultiSolverInterface
from .physics_informed_solver import (
    PINNInterface,
    PINN,
    GradientPINN,
    CausalPINN,
    CompetitivePINN,
    SelfAdaptivePINN,
    RBAPINN,
)
from .supervised_solver import (
    SupervisedSolverInterface,
    SupervisedSolver,
    ReducedOrderModelSolver,
)
from .ensemble_solver import (
    DeepEnsembleSolverInterface,
    DeepEnsembleSupervisedSolver,
    DeepEnsemblePINN,
)
from .garom import GAROM



"""Module for the Radial Field Network block."""

import torch
from torch_geometric.nn import MessagePassing
from torch_geometric.utils import remove_self_loops
from ....utils import check_positive_integer
from ....model import FeedForward


class RadialFieldNetworkBlock(MessagePassing):
    """
    Implementation of the Radial Field Network block.

    This block is used to perform message-passing between nodes and edges in a
    graph neural network, following the scheme proposed by Köhler et al. in
    2020. It serves as an inner block in a larger graph neural network
    architecture.

    The message between two nodes connected by an edge is computed by applying a
    linear transformation to the norm of the difference between the sender and
    recipient node features, together with the radial distance between the
    sender and recipient node features, followed by a non-linear activation
    function. Messages are then aggregated using an aggregation scheme
    (e.g., sum, mean, min, max, or product).

    The update step is performed by a simple addition of the incoming messages
    to the node features.

    .. seealso::

        **Original reference** Köhler, J., Klein, L., Noé, F. (2020).
        *Equivariant Flows: Exact Likelihood Generative Learning for Symmetric
        Densities*.
        In International Conference on Machine Learning.
        DOI: `<https://doi.org/10.48550/arXiv.2006.02425>`_.
    """

    def __init__(
        self,
        node_feature_dim,
        hidden_dim=64,
        n_layers=2,
        activation=torch.nn.Tanh,
        aggr="add",
        node_dim=-2,
        flow="source_to_target",
    ):
        """
        Initialization of the :class:`RadialFieldNetworkBlock` class.

        :param int node_feature_dim: The dimension of the node features.
        :param int hidden_dim: The dimension of the hidden features.
            Default is 64.
        :param int n_layers: The number of layers in the network. Default is 2.
        :param torch.nn.Module activation: The activation function.
            Default is :class:`torch.nn.Tanh`.
        :param str aggr: The aggregation scheme to use for message passing.
            Available options are "add", "mean", "min", "max", "mul".
            See :class:`torch_geometric.nn.MessagePassing` for more details.
            Default is "add".
        :param int node_dim: The axis along which to propagate. Default is -2.
        :param str flow: The direction of message passing. Available options
            are "source_to_target" and "target_to_source".
            The "source_to_target" flow means that messages are sent from
            the source node to the target node, while the "target_to_source"
            flow means that messages are sent from the target node to the
            source node. See :class:`torch_geometric.nn.MessagePassing` for more
            details. Default is "source_to_target".
        :raises AssertionError: If `node_feature_dim` is not a positive integer.
        :raises AssertionError: If `hidden_dim` is not a positive integer.
        :raises AssertionError: If `n_layers` is not a positive integer.
        """
        super().__init__(aggr=aggr, node_dim=node_dim, flow=flow)

        # Check values
        check_positive_integer(node_feature_dim, strict=True)
        check_positive_integer(hidden_dim, strict=True)
        check_positive_integer(n_layers, strict=True)

        # Layer for processing node features
        self.radial_net = FeedForward(
            input_dimensions=1,
            output_dimensions=1,
            inner_size=hidden_dim,
            n_layers=n_layers,
            func=activation,
        )

    def forward(self, x, edge_index):
        """
        Forward pass of the block, triggering the message-passing routine.

        :param x: The node features.
        :type x: torch.Tensor | LabelTensor
        :param torch.Tensor edge_index: The edge indices.
        :return: The updated node features.
        :rtype: torch.Tensor
        """
        edge_index, _ = remove_self_loops(edge_index)
        return self.propagate(edge_index=edge_index, x=x)

    def message(self, x_i, x_j):
        """
        Compute the message to be passed between nodes and edges.

        :param x_i: The node features of the recipient nodes.
        :type x_i: torch.Tensor | LabelTensor
        :param x_j: The node features of the sender nodes.
        :type x_j: torch.Tensor | LabelTensor
        :return: The message to be passed.
        :rtype: torch.Tensor
        """
        r = x_i - x_j
        return self.radial_net(torch.norm(r, dim=1, keepdim=True)) * r

    def update(self, message, x):
        """
        Update the node features with the received messages.

        :param torch.Tensor message: The message to be passed.
        :param x: The node features.
        :type x: torch.Tensor | LabelTensor
        :return: The updated node features.
        :rtype: torch.Tensor
        """
        return x + message



"""Module for the Graph Neural Operator model class."""

import torch
from torch.nn import Tanh
from .block.gno_block import GNOBlock
from .kernel_neural_operator import KernelNeuralOperator


class GraphNeuralKernel(torch.nn.Module):
    """
    Graph Neural Operator kernel model class.

    This class implements the Graph Neural Operator kernel network.

    .. seealso::

        **Original reference**: Li, Z., Kovachki, N., Azizzadenesheli, K.,
        Liu, B., Bhattacharya, K., Stuart, A., Anandkumar, A. (2020).
        *Neural Operator: Graph Kernel Network for Partial Differential
        Equations*.
        DOI: `arXiv preprint arXiv:2003.03485 <https://arxiv.org/abs/2003.03485>`_
    """

    def __init__(
        self,
        width,
        edge_features,
        n_layers=2,
        internal_n_layers=0,
        internal_layers=None,
        inner_size=None,
        internal_func=None,
        external_func=None,
        shared_weights=False,
    ):
        """
        Initialization of the :class:`GraphNeuralKernel` class.

        :param int width: The width of the kernel.
        :param int edge_features: The number of edge features.
        :param int n_layers: The number of kernel layers. Default is ``2``.
        :param int internal_n_layers: The number of layers of the neural network
            inside each kernel layer. Default is ``0``.
        :param internal_layers: The number of neurons for each layer of the
            neural network inside each kernel layer. Default is ``None``.
        :type internal_layers: list[int] | tuple[int]
        :param torch.nn.Module internal_func: The activation function used
            inside each kernel layer. If ``None``, it uses the
            :class:`torch.nn.Tanh` activation. Default is ``None``.
        :param torch.nn.Module external_func: The activation function applied to
            the output of the each kernel layer. If ``None``, it uses the
            :class:`torch.nn.Tanh` activation. Default is ``None``.
        :param bool shared_weights: If ``True``, the weights of each kernel
            layer are shared. Default is ``False``.
        """
        super().__init__()
        if external_func is None:
            external_func = Tanh
        if internal_func is None:
            internal_func = Tanh

        if shared_weights:
            self.layers = GNOBlock(
                width=width,
                edges_features=edge_features,
                n_layers=internal_n_layers,
                layers=internal_layers,
                inner_size=inner_size,
                internal_func=internal_func,
                external_func=external_func,
            )
            self.n_layers = n_layers
            self._forward_func = self._forward_shared
        else:
            self.layers = torch.nn.ModuleList(
                [
                    GNOBlock(
                        width=width,
                        edges_features=edge_features,
                        n_layers=internal_n_layers,
                        layers=internal_layers,
                        inner_size=inner_size,
                        internal_func=internal_func,
                        external_func=external_func,
                    )
                    for _ in range(n_layers)
                ]
            )
            self._forward_func = self._forward_unshared

    def _forward_unshared(self, x, edge_index, edge_attr):
        """
        Forward pass for the Graph Neural Kernel with unshared weights.

        :param x: The input tensor.
        :type x: torch.Tensor | LabelTensor
        :param torch.Tensor edge_index: The edge index.
        :param edge_attr: The edge attributes.
        :type edge_attr: torch.Tensor | LabelTensor
        :return: The output tensor.
        :rtype: torch.Tensor
        """
        for layer in self.layers:
            x = layer(x, edge_index, edge_attr)
        return x

    def _forward_shared(self, x, edge_index, edge_attr):
        """
        Forward pass for the Graph Neural Kernel with shared weights.

        :param x: The input tensor.
        :type x: torch.Tensor | LabelTensor
        :param torch.Tensor edge_index: The edge index.
        :param edge_attr: The edge attributes.
        :type edge_attr: torch.Tensor | LabelTensor
        :return: The output tensor.
        :rtype: torch.Tensor
        """
        for _ in range(self.n_layers):
            x = self.layers(x, edge_index, edge_attr)
        return x

    def forward(self, x, edge_index, edge_attr):
        """
        The forward pass of the Graph Neural Kernel.

        :param x: The input tensor.
        :type x: torch.Tensor | LabelTensor
        :param torch.Tensor edge_index: The edge index.
        :param edge_attr: The edge attributes.
        :type edge_attr: torch.Tensor | LabelTensor
        :return: The output tensor.
        :rtype: torch.Tensor
        """
        return self._forward_func(x, edge_index, edge_attr)


class GraphNeuralOperator(KernelNeuralOperator):
    """
    Graph Neural Operator model class.

    The Graph Neural Operator is a general architecture for learning operators,
    which map functions to functions. It can be trained both with Supervised
    and Physics-Informed learning strategies. The Graph Neural Operator performs
    graph convolution by means of a Graph Neural Kernel.

    .. seealso::

        **Original reference**: Li, Z., Kovachki, N., Azizzadenesheli, K.,
        Liu, B., Bhattacharya, K., Stuart, A., Anandkumar, A. (2020).
        *Neural Operator: Graph Kernel Network for Partial Differential
        Equations*.
        DOI: `arXiv preprint arXiv:2003.03485.
        <https://arxiv.org/abs/2003.03485>`_
    """

    def __init__(
        self,
        lifting_operator,
        projection_operator,
        edge_features,
        n_layers=10,
        internal_n_layers=0,
        inner_size=None,
        internal_layers=None,
        internal_func=None,
        external_func=None,
        shared_weights=True,
    ):
        """
        Initialization of the :class:`GraphNeuralOperator` class.

        :param torch.nn.Module lifting_operator: The lifting neural network
            mapping the input to its hidden dimension.
        :param torch.nn.Module projection_operator: The projection neural
            network mapping the hidden representation to the output function.
        :param int edge_features: The number of edge features.
        :param int n_layers: The number of kernel layers. Default is ``10``.
        :param int internal_n_layers: The number of layers of the neural network
            inside each kernel layer. Default is ``0``.
        :param int inner_size: The size of the hidden layers of the neural
            network inside each kernel layer. Default is ``None``.
        :param internal_layers: The number of neurons for each layer of the
            neural network inside each kernel layer. Default is ``None``.
        :type internal_layers: list[int] | tuple[int]
        :param torch.nn.Module internal_func: The activation function used
            inside each kernel layer. If ``None``, it uses the
            :class:`torch.nn.Tanh`. activation. Default is ``None``.
        :param torch.nn.Module external_func: The activation function applied to
            the output of the each kernel layer. If ``None``, it uses the
            :class:`torch.nn.Tanh`. activation. Default is ``None``.
        :param bool shared_weights: If ``True``, the weights of each kernel
            layer are shared. Default is ``False``.
        """

        if internal_func is None:
            internal_func = Tanh
        if external_func is None:
            external_func = Tanh

        super().__init__(
            lifting_operator=lifting_operator,
            integral_kernels=GraphNeuralKernel(
                width=lifting_operator.out_features,
                edge_features=edge_features,
                internal_n_layers=internal_n_layers,
                inner_size=inner_size,
                internal_layers=internal_layers,
                external_func=external_func,
                internal_func=internal_func,
                n_layers=n_layers,
                shared_weights=shared_weights,
            ),
            projection_operator=projection_operator,
        )

    def forward(self, x):
        """
        The forward pass of the Graph Neural Operator.

        :param torch_geometric.data.Batch x: The input graph.
        :return: The output tensor.
        :rtype: torch.Tensor
        """
        x, edge_index, edge_attr = x.x, x.edge_index, x.edge_attr
        x = self.lifting_operator(x)
        x = self.integral_kernels(x, edge_index, edge_attr)
        x = self.projection_operator(x)
        return x



"""Module for the Weighting Interface."""

from abc import ABCMeta, abstractmethod


class WeightingInterface(metaclass=ABCMeta):
    """
    Abstract base class for all loss weighting schemas. All weighting schemas
    should inherit from this class.
    """

    def __init__(self):
        """
        Initialization of the :class:`WeightingInterface` class.
        """
        self.condition_names = None

    @abstractmethod
    def aggregate(self, losses):
        """
        Aggregate the losses.

        :param dict losses: The dictionary of losses.
        """



"""Module for Base Continuous Convolution class."""

import torch
import warnings


class PODBlock(torch.nn.Module):
    """
    Proper Orthogonal Decomposition block.

    This block projects the input field on the proper orthogonal decomposition
    basis. Before being used, it must be fitted to the data with the ``fit``
    method, which invokes the singular value decomposition. This block is not
    trainable.

    .. note::
        All the POD modes are stored in memory, avoiding to recompute them when
        the rank changes, leading to increased memory usage.
    """

    def __init__(self, rank, scale_coefficients=True):
        """
        Initialization of the :class:`PODBlock` class.

        :param int rank: The rank of the POD layer.
        :param bool scale_coefficients: If ``True``, the coefficients are scaled
            after the projection to have zero mean and unit variance.
            Default is ``True``.
        """
        super().__init__()
        self.__scale_coefficients = scale_coefficients
        self._basis = None
        self._singular_values = None
        self._scaler = None
        self._rank = rank

    @property
    def rank(self):
        """
        The rank of the POD layer.

        :return: The rank of the POD layer.
        :rtype: int
        """
        return self._rank

    @rank.setter
    def rank(self, value):
        """
        Set the rank of the POD layer.

        :param int value: The new rank of the POD layer.
        :raises ValueError: If the rank is not a positive integer.
        """
        if value < 1 or not isinstance(value, int):
            raise ValueError("The rank must be positive integer")

        self._rank = value

    @property
    def basis(self):
        """
        The POD basis. It is a matrix whose columns are the first ``rank`` POD
        modes.

        :return: The POD basis.
        :rtype: torch.Tensor
        """
        if self._basis is None:
            return None

        return self._basis[: self.rank]

    @property
    def singular_values(self):
        """
        The singular values of the POD basis.

        :return: The singular values.
        :rtype: torch.Tensor
        """
        if self._singular_values is None:
            return None

        return self._singular_values[: self.rank]

    @property
    def scaler(self):
        """
        Return the scaler dictionary, having keys ``mean`` and ``std``
        corresponding to the mean and the standard deviation of the
        coefficients, respectively.

        :return: The scaler dictionary.
        :rtype: dict
        """
        if self._scaler is None:
            return None

        return {
            "mean": self._scaler["mean"][: self.rank],
            "std": self._scaler["std"][: self.rank],
        }

    @property
    def scale_coefficients(self):
        """
        The flag indicating if the coefficients are scaled after the projection.

        :return: The flag indicating if the coefficients are scaled.
        :rtype: bool
        """
        return self.__scale_coefficients

    def fit(self, X, randomized=True):
        """
        Set the POD basis by performing the singular value decomposition of the
        given tensor. If ``self.scale_coefficients`` is True, the coefficients
        are scaled after the projection to have zero mean and unit variance.

        :param torch.Tensor X: The input tensor to be reduced.
        """
        self._fit_pod(X, randomized)

        if self.__scale_coefficients:
            self._fit_scaler(torch.matmul(self._basis, X.T))

    def _fit_scaler(self, coeffs):
        """
        Compute the mean and the standard deviation of the given coefficients,
        which are then stored in ``self._scaler``.

        :param torch.Tensor coeffs: The coefficients to be scaled.
        """
        self._scaler = {
            "std": torch.std(coeffs, dim=1),
            "mean": torch.mean(coeffs, dim=1),
        }

    def _fit_pod(self, X, randomized):
        """
        Compute the POD basis of the given tensor, which is then stored in
        ``self._basis``.

        :param torch.Tensor X: The tensor to be reduced.
        """
        if X.device.type == "mps":  #  svd_lowrank not arailable for mps
            warnings.warn(
                "svd_lowrank not available for mps, using svd instead."
                "This may slow down computations.",
                ResourceWarning,
            )
            u, s, _ = torch.svd(X.T)
        else:
            if randomized:
                warnings.warn(
                    "Considering a randomized algorithm to compute the POD basis"
                )
                u, s, _ = torch.svd_lowrank(X.T, q=X.shape[0])

            else:
                u, s, _ = torch.svd(X.T)
        self._basis = u.T
        self._singular_values = s

    def forward(self, X):
        """
        The forward pass of the POD layer.

        :param torch.Tensor X: The input tensor to be reduced.
        :return: The reduced tensor.
        :rtype: torch.Tensor
        """
        return self.reduce(X)

    def reduce(self, X):
        """
        Reduce the input tensor to its POD representation. The POD layer must
        be fitted before being used.

        :param torch.Tensor X: The input tensor to be reduced.
        :raises RuntimeError: If the POD layer is not fitted.
        :return: The reduced tensor.
        :rtype: torch.Tensor
        """
        if self._basis is None:
            raise RuntimeError(
                "The POD layer needs to be fitted before being used."
            )

        coeff = torch.matmul(self.basis, X.T)
        if coeff.ndim == 1:
            coeff = coeff.unsqueeze(1)

        coeff = coeff.T
        if self.__scale_coefficients:
            coeff = (coeff - self.scaler["mean"]) / self.scaler["std"]

        return coeff

    def expand(self, coeff):
        """
        Expand the given coefficients to the original space. The POD layer needs
        to be fitted before being used.

        :param torch.Tensor coeff: The coefficients to be expanded.
        :raises RuntimeError: If the POD layer is not fitted.
        :return: The expanded tensor.
        :rtype: torch.Tensor
        """
        if self._basis is None:
            raise RuntimeError(
                "The POD layer needs to be trained before being used."
            )

        if self.__scale_coefficients:
            coeff = coeff * self.scaler["std"] + self.scaler["mean"]
        predicted = torch.matmul(self.basis.T, coeff.T).T

        if predicted.ndim == 1:
            predicted = predicted.unsqueeze(0)

        return predicted



"""Module for the DomainEquationCondition class."""

from .condition_interface import ConditionInterface
from ..utils import check_consistency
from ..domain import DomainInterface
from ..equation.equation_interface import EquationInterface


class DomainEquationCondition(ConditionInterface):
    """
    Condition defined by a domain and an equation. It can be used in Physics
    Informed problems. Before using this condition, make sure that input data
    are correctly sampled from the domain.
    """

    __slots__ = ["domain", "equation"]

    def __init__(self, domain, equation):
        """
        Initialise the object by storing the domain and equation.

        :param DomainInterface domain: Domain object containing the domain data.
        :param EquationInterface equation: Equation object containing the
            equation data.
        """
        super().__init__()
        self.domain = domain
        self.equation = equation

    def __setattr__(self, key, value):
        if key == "domain":
            check_consistency(value, (DomainInterface, str))
            DomainEquationCondition.__dict__[key].__set__(self, value)
        elif key == "equation":
            check_consistency(value, (EquationInterface))
            DomainEquationCondition.__dict__[key].__set__(self, value)
        elif key in ("_problem"):
            super().__setattr__(key, value)



"""Module for the Fourier Neural Operator model class."""

import warnings
import torch
from torch import nn
from ..label_tensor import LabelTensor
from ..utils import check_consistency
from .block.fourier_block import FourierBlock1D, FourierBlock2D, FourierBlock3D
from .kernel_neural_operator import KernelNeuralOperator


class FourierIntegralKernel(torch.nn.Module):
    """
    Fourier Integral Kernel model class.

    This class implements the Fourier Integral Kernel network, which
    performs global convolution in the Fourier space.

    .. seealso::

        **Original reference**: Li, Z., Kovachki, N., Azizzadenesheli, K., Liu,
        B., Bhattacharya, K., Stuart, A., & Anandkumar, A. (2020).
        *Fourier neural operator for parametric partial differential equations*.
        DOI: `arXiv preprint arXiv:2010.08895.
        <https://arxiv.org/abs/2010.08895>`_
    """

    def __init__(
        self,
        input_numb_fields,
        output_numb_fields,
        n_modes,
        dimensions=3,
        padding=8,
        padding_type="constant",
        inner_size=20,
        n_layers=2,
        func=nn.Tanh,
        layers=None,
    ):
        """
        Initialization of the :class:`FourierIntegralKernel` class.

        :param int input_numb_fields: The number of input fields.
        :param int output_numb_fields: The number of output fields.
        :param n_modes: The number of modes.
        :type n_modes: int | list[int]
        :param int dimensions: The number of dimensions. It can be set to ``1``,
            ``2``, or ``3``. Default is ``3``.
        :param int padding: The padding size. Default is ``8``.
        :param str padding_type: The padding strategy. Default is ``constant``.
        :param int inner_size: The inner size. Default is ``20``.
        :param int n_layers: The number of layers. Default is ``2``.
        :param func: The activation function. If a list is passed, it must have
            the same length as ``n_layers``. If a single function is passed, it
            is used for all layers, except for the last one.
            Default is :class:`torch.nn.Tanh`.
        :type func: torch.nn.Module | list[torch.nn.Module]
        :param list[int] layers: The list of the dimension of inner layers.
            If ``None``, ``n_layers`` of dimension ``inner_size`` are used.
            Otherwise, it overrides the values passed to ``n_layers`` and
            ``inner_size``. Default is ``None``.
        :raises RuntimeError: If the number of layers and functions are
            inconsistent.
        :raises RunTimeError: If the number of layers and modes are
            inconsistent.
        """
        super().__init__()

        # check type consistency
        self._check_consistency(
            dimensions,
            padding,
            padding_type,
            inner_size,
            n_layers,
            func,
            layers,
            n_modes,
        )

        # assign padding
        self._padding = padding

        # initialize fourier layer for each dimension
        fourier_layer = self._get_fourier_block(dimensions)

        # Here we build the FNO kernels by stacking Fourier Blocks

        # 1. Assign output dimensions for each FNO layer
        if layers is None:
            layers = [inner_size] * n_layers

        # 2. Assign activation functions for each FNO layer
        if isinstance(func, list):
            if len(layers) != len(func):
                raise RuntimeError(
                    "Inconsistent number of layers and functions."
                )
            _functions = func
        else:
            _functions = [func for _ in range(len(layers) - 1)]
        _functions.append(torch.nn.Identity)

        # 3. Assign modes functions for each FNO layer
        if isinstance(n_modes, list):
            if all(isinstance(i, list) for i in n_modes) and len(layers) != len(
                n_modes
            ):
                raise RuntimeError("Inconsistent number of layers and modes.")
            if all(isinstance(i, int) for i in n_modes):
                n_modes = [n_modes] * len(layers)
        else:
            n_modes = [n_modes] * len(layers)

        # 4. Build the FNO network
        tmp_layers = [input_numb_fields] + layers + [output_numb_fields]
        self._layers = nn.Sequential(
            *[
                fourier_layer(
                    input_numb_fields=tmp_layers[i],
                    output_numb_fields=tmp_layers[i + 1],
                    n_modes=n_modes[i],
                    activation=_functions[i],
                )
                for i in range(len(layers))
            ]
        )

        # 5. Padding values for spectral conv
        if isinstance(padding, int):
            padding = [padding] * dimensions
        self._ipad = [-pad if pad > 0 else None for pad in padding[:dimensions]]
        self._padding_type = padding_type
        self._pad = [
            val for pair in zip([0] * dimensions, padding) for val in pair
        ]

    def forward(self, x):
        """
        Forward pass for the :class:`FourierIntegralKernel` model.

        :param x: The input tensor for performing the computation. Depending
            on the ``dimensions`` in the initialization, it expects a tensor
            with the following shapes:
            * 1D tensors: ``[batch, X, channels]``
            * 2D tensors: ``[batch, X, Y, channels]``
            * 3D tensors: ``[batch, X, Y, Z, channels]``
        :type x: torch.Tensor | LabelTensor
        :raises Warning: If a LabelTensor is passed as input.
        :return: The output tensor.
        :rtype: torch.Tensor
        """
        if isinstance(x, LabelTensor):
            warnings.warn(
                "LabelTensor passed as input is not allowed,"
                " casting LabelTensor to Torch.Tensor"
            )
            x = x.as_subclass(torch.Tensor)
        # permuting the input [batch, channels, x, y, ...]
        permutation_idx = [0, x.ndim - 1, *list(range(1, x.ndim - 1))]
        x = x.permute(permutation_idx)

        # padding the input
        x = torch.nn.functional.pad(x, pad=self._pad, mode=self._padding_type)

        # apply fourier layers
        x = self._layers(x)

        # remove padding
        idxs = [slice(None), slice(None)] + [slice(pad) for pad in self._ipad]
        x = x[idxs]

        # permuting back [batch, x, y, ..., channels]
        permutation_idx = [0, *list(range(2, x.ndim)), 1]
        x = x.permute(permutation_idx)

        return x

    @staticmethod
    def _check_consistency(
        dimensions,
        padding,
        padding_type,
        inner_size,
        n_layers,
        func,
        layers,
        n_modes,
    ):
        """
        Check the consistency of the input parameters.


        :param int dimensions: The number of dimensions.
        :param int padding: The padding size.
        :param str padding_type: The padding strategy.
        :param int inner_size: The inner size.
        :param int n_layers: The number of layers.
        :param func: The activation function.
        :type func: torch.nn.Module | list[torch.nn.Module]
        :param list[int] layers: The list of the dimension of inner layers.
        :param n_modes: The number of modes.
        :type n_modes: int | list[int]
        :raises ValueError: If the input is not consistent.
        """
        check_consistency(dimensions, int)
        check_consistency(padding, int)
        check_consistency(padding_type, str)
        check_consistency(inner_size, int)
        check_consistency(n_layers, int)
        check_consistency(func, nn.Module, subclass=True)

        if layers is not None:
            if isinstance(layers, (tuple, list)):
                check_consistency(layers, int)
            else:
                raise ValueError("layers must be tuple or list of int.")
        if not isinstance(n_modes, (list, tuple, int)):
            raise ValueError(
                "n_modes must be a int or list or tuple of valid modes."
                " More information on the official documentation."
            )

    @staticmethod
    def _get_fourier_block(dimensions):
        """
        Retrieve the Fourier Block class based on the number of dimensions.

        :param int dimensions: The number of dimensions.
        :raises NotImplementedError: If the number of dimensions is not 1, 2,
            or 3.
        :return: The Fourier Block class.
        :rtype: FourierBlock1D | FourierBlock2D | FourierBlock3D
        """
        if dimensions == 1:
            return FourierBlock1D
        if dimensions == 2:
            return FourierBlock2D
        if dimensions == 3:
            return FourierBlock3D
        raise NotImplementedError("FNO implemented only for 1D/2D/3D data.")


class FNO(KernelNeuralOperator):
    """
    Fourier Neural Operator model class.

    The Fourier Neural Operator (FNO) is a general architecture for learning
    operators, which  map functions to functions. It can be trained both with
    Supervised and Physics_Informed learning strategies. The Fourier Neural
    Operator performs global convolution in the Fourier space.

    .. seealso::

        **Original reference**: Li, Z., Kovachki, N., Azizzadenesheli, K.,
        Liu, B., Bhattacharya, K., Stuart, A., & Anandkumar, A. (2020).
        *Fourier neural operator for parametric partial differential equations*.
        DOI: `arXiv preprint arXiv:2010.08895.
        <https://arxiv.org/abs/2010.08895>`_
    """

    def __init__(
        self,
        lifting_net,
        projecting_net,
        n_modes,
        dimensions=3,
        padding=8,
        padding_type="constant",
        inner_size=20,
        n_layers=2,
        func=nn.Tanh,
        layers=None,
    ):
        """
        :param torch.nn.Module lifting_net: The lifting neural network mapping
            the input to its hidden dimension.
        :param torch.nn.Module projecting_net: The projection neural network
            mapping the hidden representation to the output function.
        :param n_modes: The number of modes.
        :type n_modes: int | list[int]
        :param int dimensions: The number of dimensions. It can be set to ``1``,
            ``2``, or ``3``. Default is ``3``.
        :param int padding: The padding size. Default is ``8``.
        :param str padding_type: The padding strategy. Default is ``constant``.
        :param int inner_size: The inner size. Default is ``20``.
        :param int n_layers: The number of layers. Default is ``2``.
        :param func: The activation function. If a list is passed, it must have
            the same length as ``n_layers``. If a single function is passed, it
            is used for all layers, except for the last one.
            Default is :class:`torch.nn.Tanh`.
        :type func: torch.nn.Module | list[torch.nn.Module]
        :param list[int] layers: The list of the dimension of inner layers.
            If ``None``, ``n_layers`` of dimension ``inner_size`` are used.
            Otherwise, it overrides the values passed to ``n_layers`` and
            ``inner_size``. Default is ``None``.
        """
        lifting_operator_out = lifting_net(
            torch.rand(size=next(lifting_net.parameters()).size())
        ).shape[-1]
        super().__init__(
            lifting_operator=lifting_net,
            projection_operator=projecting_net,
            integral_kernels=FourierIntegralKernel(
                input_numb_fields=lifting_operator_out,
                output_numb_fields=next(projecting_net.parameters()).size(),
                n_modes=n_modes,
                dimensions=dimensions,
                padding=padding,
                padding_type=padding_type,
                inner_size=inner_size,
                n_layers=n_layers,
                func=func,
                layers=layers,
            ),
        )

    def forward(self, x):
        """
        Forward pass for the :class:`FourierNeuralOperator` model.

        The ``lifting_net`` maps the input to the hidden dimension.
        Then, several layers of Fourier blocks are applied. Finally, the
        ``projection_net`` maps the hidden representation to the output
        function.

        :param x: The input tensor for performing the computation. Depending
            on the ``dimensions`` in the initialization, it expects a tensor
            with the following shapes:

            * 1D tensors: ``[batch, X, channels]``
            * 2D tensors: ``[batch, X, Y, channels]``
            * 3D tensors: ``[batch, X, Y, Z, channels]``

        :type x: torch.Tensor | LabelTensor
        :return: The output tensor.
        :rtype: torch.Tensor
        """

        if isinstance(x, LabelTensor):
            x = x.as_subclass(torch.Tensor)
        return super().forward(x)



"""Module for utility functions."""

import types
from functools import reduce
import torch

from .label_tensor import LabelTensor


# Codacy error unused parameters
def custom_warning_format(
    message, category, filename, lineno, file=None, line=None
):
    """
    Custom warning formatting function.

    :param str message: The warning message.
    :param Warning category: The warning category.
    :param str filename: The filename where the warning is raised.
    :param int lineno: The line number where the warning is raised.
    :param str file: The file object where the warning is raised.
        Default is None.
    :param int line: The line where the warning is raised.
    :return: The formatted warning message.
    :rtype: str
    """
    return f"{filename}: {category.__name__}: {message}\n"


def check_consistency(object_, object_instance, subclass=False):
    """
    Check if an object maintains inheritance consistency.

    This function checks whether a given object is an instance of a specified
    class or, if ``subclass=True``, whether it is a subclass of the specified
    class.

    :param object: The object to check.
    :type object: Iterable | Object
    :param Object object_instance: The expected parent class.
    :param bool subclass: If True, checks whether ``object_`` is a subclass
        of ``object_instance`` instead of an instance. Default is ``False``.
    :raises ValueError: If ``object_`` does not inherit from ``object_instance``
        as expected.
    """
    if not isinstance(object_, (list, set, tuple)):
        object_ = [object_]

    for obj in object_:
        try:
            if not subclass:
                assert isinstance(obj, object_instance)
            else:
                assert issubclass(obj, object_instance)
        except AssertionError as e:
            raise ValueError(
                f"{type(obj).__name__} must be {object_instance}."
            ) from e


def labelize_forward(forward, input_variables, output_variables):
    """
    Decorator to enable or disable the use of
    :class:`~pina.label_tensor.LabelTensor` during the forward pass.

    :param Callable forward: The forward function of a :class:`torch.nn.Module`.
    :param list[str] input_variables: The names of the input variables of a
        :class:`~pina.problem.abstract_problem.AbstractProblem`.
    :param list[str] output_variables: The names of the output variables of a
        :class:`~pina.problem.abstract_problem.AbstractProblem`.
    :return: The decorated forward function.
    :rtype: Callable
    """

    def wrapper(x, *args, **kwargs):
        """
        Decorated forward function.

        :param LabelTensor x: The labelized input of the forward pass of an
            instance of :class:`torch.nn.Module`.
        :param Iterable args: Additional positional arguments passed to
            ``forward`` method.
        :param dict kwargs: Additional keyword arguments passed to
            ``forward`` method.
        :return: The labelized output of the forward pass of an instance of
            :class:`torch.nn.Module`.
        :rtype: LabelTensor
        """
        x = x.extract(input_variables)
        output = forward(x, *args, **kwargs)
        # keep it like this, directly using LabelTensor(...) raises errors
        # when compiling the code
        output = output.as_subclass(LabelTensor)
        output.labels = output_variables
        return output

    return wrapper


def merge_tensors(tensors):
    """
    Merge a list of :class:`~pina.label_tensor.LabelTensor` instances into a
    single :class:`~pina.label_tensor.LabelTensor` tensor, by applying
    iteratively the cartesian product.

    :param list[LabelTensor] tensors: The list of tensors to merge.
    :raises ValueError: If the list of tensors is empty.
    :return: The merged tensor.
    :rtype: LabelTensor
    """
    if tensors:
        return reduce(merge_two_tensors, tensors[1:], tensors[0])
    raise ValueError("Expected at least one tensor")


def merge_two_tensors(tensor1, tensor2):
    """
    Merge two :class:`~pina.label_tensor.LabelTensor` instances into a single
    :class:`~pina.label_tensor.LabelTensor` tensor, by applying the cartesian
    product.

    :param LabelTensor tensor1: The first tensor to merge.
    :param LabelTensor tensor2: The second tensor to merge.
    :return: The merged tensor.
    :rtype: LabelTensor
    """
    n1 = tensor1.shape[0]
    n2 = tensor2.shape[0]

    tensor1 = LabelTensor(tensor1.repeat(n2, 1), labels=tensor1.labels)
    tensor2 = LabelTensor(
        tensor2.repeat_interleave(n1, dim=0), labels=tensor2.labels
    )
    return tensor1.append(tensor2)


def torch_lhs(n, dim):
    """
    The Latin Hypercube Sampling torch routine, sampling in :math:`[0, 1)`$.

    :param int n: The number of points to sample.
    :param int dim: The number of dimensions of the sampling space.
    :raises TypeError: If `n` or `dim` are not integers.
    :raises ValueError: If `dim` is less than 1.
    :return: The sampled points.
    :rtype: torch.tensor
    """

    if not isinstance(n, int):
        raise TypeError("number of point n must be int")

    if not isinstance(dim, int):
        raise TypeError("dim must be int")

    if dim < 1:
        raise ValueError("dim must be greater than one")

    samples = torch.rand(size=(n, dim))

    perms = torch.tile(torch.arange(1, n + 1), (dim, 1))

    for row in range(dim):
        idx_perm = torch.randperm(perms.shape[-1])
        perms[row, :] = perms[row, idx_perm]

    perms = perms.T

    samples = (perms - samples) / n

    return samples


def is_function(f):
    """
    Check if the given object is a function or a lambda.

    :param Object f: The object to be checked.
    :return: ``True`` if ``f`` is a function, ``False`` otherwise.
    :rtype: bool
    """
    return isinstance(f, (types.FunctionType, types.LambdaType))


def chebyshev_roots(n):
    """
    Compute the roots of the Chebyshev polynomial of degree ``n``.

    :param int n: The number of roots to return.
    :return: The roots of the Chebyshev polynomials.
    :rtype: torch.Tensor
    """
    pi = torch.acos(torch.zeros(1)).item() * 2
    k = torch.arange(n)
    nodes = torch.sort(torch.cos(pi * (k + 0.5) / n))[0]
    return nodes


def check_positive_integer(value, strict=True):
    """
    Check if the value is a positive integer.

    :param int value: The value to check.
    :param bool strict: If True, the value must be strictly positive.
        Default is True.
    :raises AssertionError: If the value is not a positive integer.
    """
    if strict:
        assert (
            isinstance(value, int) and value > 0
        ), f"Expected a strictly positive integer, got {value}."
    else:
        assert (
            isinstance(value, int) and value >= 0
        ), f"Expected a non-negative integer, got {value}."



"""Module for the GAROM solver."""

import torch
from torch.nn.modules.loss import _Loss
from .solver import MultiSolverInterface
from ..condition import InputTargetCondition
from ..utils import check_consistency
from ..loss import LossInterface, PowerLoss


class GAROM(MultiSolverInterface):
    """
    GAROM solver class. This class implements Generative Adversarial Reduced
    Order Model solver, using user specified ``models`` to solve a specific
    order reduction ``problem``.

    .. seealso::

        **Original reference**: Coscia, D., Demo, N., & Rozza, G. (2023).
        *Generative Adversarial Reduced Order Modelling*.
        DOI: `arXiv preprint arXiv:2305.15881.
        <https://doi.org/10.48550/arXiv.2305.15881>`_.
    """

    accepted_conditions_types = InputTargetCondition

    def __init__(
        self,
        problem,
        generator,
        discriminator,
        loss=None,
        optimizer_generator=None,
        optimizer_discriminator=None,
        scheduler_generator=None,
        scheduler_discriminator=None,
        gamma=0.3,
        lambda_k=0.001,
        regularizer=False,
    ):
        """
        Initialization of the :class:`GAROM` class.

        :param AbstractProblem problem: The formulation of the problem.
        :param torch.nn.Module generator: The generator model.
        :param torch.nn.Module discriminator: The discriminator model.
        :param torch.nn.Module loss: The loss function to be minimized.
            If ``None``, :class:`~pina.loss.power_loss.PowerLoss` with ``p=1``
            is used. Default is ``None``.
        :param Optimizer optimizer_generator: The optimizer for the generator.
            If ``None``, the :class:`torch.optim.Adam` optimizer is used.
            Default is ``None``.
        :param Optimizer optimizer_discriminator: The optimizer for the
            discriminator. If ``None``, the :class:`torch.optim.Adam`
            optimizer is used. Default is ``None``.
        :param Scheduler scheduler_generator: The learning rate scheduler for
            the generator.
            If ``None``, the :class:`torch.optim.lr_scheduler.ConstantLR`
            scheduler is used. Default is ``None``.
        :param Scheduler scheduler_discriminator: The learning rate scheduler
            for the discriminator.
            If ``None``, the :class:`torch.optim.lr_scheduler.ConstantLR`
            scheduler is used. Default is ``None``.
        :param float gamma: Ratio of expected loss for generator and
            discriminator. Default is ``0.3``.
        :param float lambda_k: Learning rate for control theory optimization.
            Default is ``0.001``.
        :param bool regularizer: If ``True``, uses a regularization term in the
            GAROM loss. Default is ``False``.
        """

        # set loss
        if loss is None:
            loss = PowerLoss(p=1)

        super().__init__(
            models=[generator, discriminator],
            problem=problem,
            optimizers=[optimizer_generator, optimizer_discriminator],
            schedulers=[
                scheduler_generator,
                scheduler_discriminator,
            ],
            use_lt=False,
        )

        # check consistency
        check_consistency(
            loss, (LossInterface, _Loss, torch.nn.Module), subclass=False
        )
        self._loss_fn = loss

        # set automatic optimization for GANs
        self.automatic_optimization = False

        # check consistency
        check_consistency(gamma, float)
        check_consistency(lambda_k, float)
        check_consistency(regularizer, bool)

        # began hyperparameters
        self.k = 0
        self.gamma = gamma
        self.lambda_k = lambda_k
        self.regularizer = float(regularizer)

    def forward(self, x, mc_steps=20, variance=False):
        """
        Forward pass implementation.

        :param torch.Tensor x: The input tensor.
        :param int mc_steps: Number of Montecarlo samples to approximate the
            expected value. Default is ``20``.
        :param bool variance: If ``True``, the method returns also the variance
            of the solution. Default is ``False``.
        :return: The expected value of the generator distribution. If
            ``variance=True``, the method returns also the variance.
        :rtype: torch.Tensor | tuple[torch.Tensor, torch.Tensor]
        """

        # sampling
        field_sample = [self.sample(x) for _ in range(mc_steps)]
        field_sample = torch.stack(field_sample)

        # extract mean
        mean = field_sample.mean(dim=0)

        if variance:
            var = field_sample.var(dim=0)
            return mean, var

        return mean

    def sample(self, x):
        """
        Sample from the generator distribution.

        :param torch.Tensor x: The input tensor.
        :return: The generated sample.
        :rtype: torch.Tensor
        """
        # sampling
        return self.generator(x)

    def _train_generator(self, parameters, snapshots):
        """
        Train the generator model.

        :param torch.Tensor parameters: The input tensor.
        :param torch.Tensor snapshots: The target tensor.
        :return: The residual loss and the generator loss.
        :rtype: tuple[torch.Tensor, torch.Tensor]
        """
        self.optimizer_generator.instance.zero_grad()

        # Generate a batch of images
        generated_snapshots = self.sample(parameters)

        # generator loss
        r_loss = self._loss_fn(snapshots, generated_snapshots)
        d_fake = self.discriminator([generated_snapshots, parameters])
        g_loss = (
            self._loss_fn(d_fake, generated_snapshots)
            + self.regularizer * r_loss
        )

        # backward step
        g_loss.backward()
        self.optimizer_generator.instance.step()
        self.scheduler_generator.instance.step()

        return r_loss, g_loss

    def _train_discriminator(self, parameters, snapshots):
        """
        Train the discriminator model.

        :param torch.Tensor parameters: The input tensor.
        :param torch.Tensor snapshots: The target tensor.
        :return: The residual loss and the generator loss.
        :rtype: tuple[torch.Tensor, torch.Tensor]
        """
        self.optimizer_discriminator.instance.zero_grad()

        # Generate a batch of images
        generated_snapshots = self.sample(parameters)

        # Discriminator pass
        d_real = self.discriminator([snapshots, parameters])
        d_fake = self.discriminator([generated_snapshots, parameters])

        # evaluate loss
        d_loss_real = self._loss_fn(d_real, snapshots)
        d_loss_fake = self._loss_fn(d_fake, generated_snapshots.detach())
        d_loss = d_loss_real - self.k * d_loss_fake

        # backward step
        d_loss.backward()
        self.optimizer_discriminator.instance.step()
        self.scheduler_discriminator.instance.step()

        return d_loss_real, d_loss_fake, d_loss

    def _update_weights(self, d_loss_real, d_loss_fake):
        """
        Update the weights of the generator and discriminator models.

        :param torch.Tensor d_loss_real: The discriminator loss computed on
            dataset samples.
        :param torch.Tensor d_loss_fake: The discriminator loss computed on
            generated samples.
        :return: The difference between the loss computed on the dataset samples
            and the loss computed on the generated samples.
        :rtype: torch.Tensor
        """

        diff = torch.mean(self.gamma * d_loss_real - d_loss_fake)

        # Update weight term for fake samples
        self.k += self.lambda_k * diff.item()
        self.k = min(max(self.k, 0), 1)  # Constraint to interval [0, 1]
        return diff

    def optimization_cycle(self, batch):
        """
        The optimization cycle for the GAROM solver.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :return: The losses computed for all conditions in the batch, casted
            to a subclass of :class:`torch.Tensor`. It should return a dict
            containing the condition name and the associated scalar loss.
        :rtype: dict
        """
        condition_loss = {}
        for condition_name, points in batch:
            parameters, snapshots = (
                points["input"],
                points["target"],
            )
            d_loss_real, d_loss_fake, d_loss = self._train_discriminator(
                parameters, snapshots
            )
            r_loss, g_loss = self._train_generator(parameters, snapshots)
            diff = self._update_weights(d_loss_real, d_loss_fake)
            condition_loss[condition_name] = r_loss

        # some extra logging
        self.store_log("d_loss", float(d_loss), self.get_batch_size(batch))
        self.store_log("g_loss", float(g_loss), self.get_batch_size(batch))
        self.store_log(
            "stability_metric",
            float(d_loss_real + torch.abs(diff)),
            self.get_batch_size(batch),
        )
        return condition_loss

    def validation_step(self, batch):
        """
        The validation step for the PINN solver.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :return: The loss of the validation step.
        :rtype: torch.Tensor
        """
        condition_loss = {}
        for condition_name, points in batch:
            parameters, snapshots = (
                points["input"],
                points["target"],
            )
            snapshots_gen = self.generator(parameters)
            condition_loss[condition_name] = self._loss_fn(
                snapshots, snapshots_gen
            )
        loss = self.weighting.aggregate(condition_loss)
        self.store_log("val_loss", loss, self.get_batch_size(batch))
        return loss

    def test_step(self, batch):
        """
        The test step for the PINN solver.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :return: The loss of the test step.
        :rtype: torch.Tensor
        """
        condition_loss = {}
        for condition_name, points in batch:
            parameters, snapshots = (
                points["input"],
                points["target"],
            )
            snapshots_gen = self.generator(parameters)
            condition_loss[condition_name] = self._loss_fn(
                snapshots, snapshots_gen
            )
        loss = self.weighting.aggregate(condition_loss)
        self.store_log("test_loss", loss, self.get_batch_size(batch))
        return loss

    @property
    def generator(self):
        """
        The generator model.

        :return: The generator model.
        :rtype: torch.nn.Module
        """
        return self.models[0]

    @property
    def discriminator(self):
        """
        The discriminator model.

        :return: The discriminator model.
        :rtype: torch.nn.Module
        """
        return self.models[1]

    @property
    def optimizer_generator(self):
        """
        The optimizer for the generator.

        :return: The optimizer for the generator.
        :rtype: Optimizer
        """
        return self.optimizers[0]

    @property
    def optimizer_discriminator(self):
        """
        The optimizer for the discriminator.

        :return: The optimizer for the discriminator.
        :rtype: Optimizer
        """
        return self.optimizers[1]

    @property
    def scheduler_generator(self):
        """
        The scheduler for the generator.

        :return: The scheduler for the generator.
        :rtype: Scheduler
        """
        return self.schedulers[0]

    @property
    def scheduler_discriminator(self):
        """
        The scheduler for the discriminator.

        :return: The scheduler for the discriminator.
        :rtype: Scheduler
        """
        return self.schedulers[1]



"""Module to perform integration for continuous convolution."""

import torch


class Integral:
    """
    Class allowing integration for continous convolution.
    """

    def __init__(self, param):
        """
        Initializzation of the :class:`Integral` class.

        :param param: The type of continuous convolution.
        :type param: string
        :raises TypeError: If the parameter is neither ``discrete``
            nor ``continuous``.
        """
        if param == "discrete":
            self.make_integral = self.integral_param_disc
        elif param == "continuous":
            self.make_integral = self.integral_param_cont
        else:
            raise TypeError

    def __call__(self, *args, **kwds):
        """
        Call the integral function

        :param list args: Arguments for the integral function.
        :param dict kwds: Keyword arguments for the integral function.
        :return: The integral of the input.
        :rtype: torch.tensor
        """
        return self.make_integral(*args, **kwds)

    def _prepend_zero(self, x):
        """
        Create bins to perform integration.

        :param torch.Tensor x: The input tensor.
        :return: The bins for the integral.
        :rtype: torch.Tensor
        """
        return torch.cat((torch.zeros(1, dtype=x.dtype, device=x.device), x))

    def integral_param_disc(self, x, y, idx):
        """
        Perform discrete integration with discrete parameters.

        :param torch.Tensor x: The first input tensor.
        :param torch.Tensor y: The second input tensor.
        :param list[int] idx: The indices for different strides.
        :return: The discrete integral.
        :rtype: torch.Tensor
        """
        cs_idxes = self._prepend_zero(torch.cumsum(torch.tensor(idx), 0))
        cs = self._prepend_zero(torch.cumsum(x.flatten() * y.flatten(), 0))
        return cs[cs_idxes[1:]] - cs[cs_idxes[:-1]]

    def integral_param_cont(self, x, y, idx):
        """
        Perform continuous integration with continuous parameters.

        :param torch.Tensor x: The first input tensor.
        :param torch.Tensor y: The second input tensor.
        :param list[int] idx: The indices for different strides.
        :raises NotImplementedError: The method is not implemented.
        """
        raise NotImplementedError



"""Formulation of the diffusion-reaction problem."""

import torch
from ... import Condition
from ...domain import CartesianDomain
from ...operator import grad, laplacian
from ...equation import Equation, FixedValue
from ...problem import SpatialProblem, TimeDependentProblem


def diffusion_reaction(input_, output_):
    """
    Implementation of the diffusion-reaction equation.

    :param LabelTensor input_: Input data of the problem.
    :param LabelTensor output_: Output data of the problem.
    :return: The residual of the diffusion-reaction equation.
    :rtype: LabelTensor
    """
    x = input_.extract("x")
    t = input_.extract("t")
    u_t = grad(output_, input_, components=["u"], d=["t"])
    u_xx = laplacian(output_, input_, components=["u"], d=["x"])
    r = torch.exp(-t) * (
        1.5 * torch.sin(2 * x)
        + (8 / 3) * torch.sin(3 * x)
        + (15 / 4) * torch.sin(4 * x)
        + (63 / 8) * torch.sin(8 * x)
    )
    return u_t - u_xx - r


def initial_condition(input_, output_):
    """
    Definition of the initial condition of the diffusion-reaction problem.

    :param LabelTensor input_: Input data of the problem.
    :param LabelTensor output_: Output data of the problem.
    :return: The residual of the initial condition.
    :rtype: LabelTensor
    """
    x = input_.extract("x")
    u_0 = (
        torch.sin(x)
        + (1 / 2) * torch.sin(2 * x)
        + (1 / 3) * torch.sin(3 * x)
        + (1 / 4) * torch.sin(4 * x)
        + (1 / 8) * torch.sin(8 * x)
    )
    return output_ - u_0


class DiffusionReactionProblem(TimeDependentProblem, SpatialProblem):
    r"""
    Implementation of the diffusion-reaction problem in the spatial interval
    :math:`[-\pi, \pi]` and temporal interval :math:`[0, 1]`.

    .. seealso::
        **Original reference**: Si, Chenhao, et al. *Complex Physics-Informed
        Neural Network.* arXiv preprint arXiv:2502.04917 (2025).
        DOI: `arXiv:2502.04917 <https://arxiv.org/abs/2502.04917>`_.

    :Example:
        >>> problem = DiffusionReactionProblem()
    """

    output_variables = ["u"]
    spatial_domain = CartesianDomain({"x": [-torch.pi, torch.pi]})
    temporal_domain = CartesianDomain({"t": [0, 1]})

    domains = {
        "D": CartesianDomain({"x": [-torch.pi, torch.pi], "t": [0, 1]}),
        "g1": CartesianDomain({"x": -torch.pi, "t": [0, 1]}),
        "g2": CartesianDomain({"x": torch.pi, "t": [0, 1]}),
        "t0": CartesianDomain({"x": [-torch.pi, torch.pi], "t": 0.0}),
    }

    conditions = {
        "D": Condition(domain="D", equation=Equation(diffusion_reaction)),
        "g1": Condition(domain="g1", equation=FixedValue(0.0)),
        "g2": Condition(domain="g2", equation=FixedValue(0.0)),
        "t0": Condition(domain="t0", equation=Equation(initial_condition)),
    }

    def solution(self, pts):
        """
        Implementation of the analytical solution of the diffusion-reaction
        problem.

        :param LabelTensor pts: Points where the solution is evaluated.
        :return: The analytical solution of the diffusion-reaction problem.
        :rtype: LabelTensor
        """
        t = pts.extract("t")
        x = pts.extract("x")
        sol = torch.exp(-t) * (
            torch.sin(x)
            + (1 / 2) * torch.sin(2 * x)
            + (1 / 3) * torch.sin(3 * x)
            + (1 / 4) * torch.sin(4 * x)
            + (1 / 8) * torch.sin(8 * x)
        )
        sol.labels = self.output_variables
        return sol



"""Old module for adaptive functions. Deprecated in 0.2.0."""

import warnings

from ..adaptive_function import *
from ..utils import custom_warning_format

# back-compatibility 0.1
# Set the custom format for warnings
warnings.formatwarning = custom_warning_format
warnings.filterwarnings("always", category=DeprecationWarning)
warnings.warn(
    "'pina.adaptive_functions' is deprecated and will be removed "
    "in future versions. Please use 'pina.adaptive_function' instead.",
    DeprecationWarning,
)



"""Module for the PINA Torch Optimizer"""

import torch

from ..utils import check_consistency
from .optimizer_interface import Optimizer


class TorchOptimizer(Optimizer):
    """
    A wrapper class for using PyTorch optimizers.
    """

    def __init__(self, optimizer_class, **kwargs):
        """
        Initialization of the :class:`TorchOptimizer` class.

        :param torch.optim.Optimizer optimizer_class: A
            :class:`torch.optim.Optimizer` class.
        :param dict kwargs: Additional parameters passed to ``optimizer_class``,
            see more
            `here <https://pytorch.org/docs/stable/optim.html#algorithms>`_.
        """
        check_consistency(optimizer_class, torch.optim.Optimizer, subclass=True)

        self.optimizer_class = optimizer_class
        self.kwargs = kwargs
        self._optimizer_instance = None

    def hook(self, parameters):
        """
        Initialize the optimizer instance with the given parameters.

        :param dict parameters: The parameters of the model to be optimized.
        """
        self._optimizer_instance = self.optimizer_class(
            parameters, **self.kwargs
        )

    @property
    def instance(self):
        """
        Get the optimizer instance.

        :return: The optimizer instance.
        :rtype: torch.optim.Optimizer
        """
        return self._optimizer_instance



"""Module for loss functions and weighting functions."""

__all__ = [
    "LossInterface",
    "LpLoss",
    "PowerLoss",
    "WeightingInterface",
    "ScalarWeighting",
    "NeuralTangentKernelWeighting",
]

from .loss_interface import LossInterface
from .power_loss import PowerLoss
from .lp_loss import LpLoss
from .weighting_interface import WeightingInterface
from .scalar_weighting import ScalarWeighting
from .ntk_weighting import NeuralTangentKernelWeighting



"""Module for the PINA dataset classes."""

from abc import abstractmethod, ABC
from torch.utils.data import Dataset
from torch_geometric.data import Data
from ..graph import Graph, LabelBatch


class PinaDatasetFactory:
    """
    Factory class for the PINA dataset.

    Depending on the data type inside the conditions, it instanciate an object
    belonging to the appropriate subclass of 
    :class:`~pina.data.dataset.PinaDataset`. The possible subclasses are:

    - :class:`~pina.data.dataset.PinaTensorDataset`, for handling \
        :class:`torch.Tensor` and :class:`~pina.label_tensor.LabelTensor` data.
    - :class:`~pina.data.dataset.PinaGraphDataset`, for handling \
        :class:`~pina.graph.Graph` and :class:`~torch_geometric.data.Data` data.
    """

    def __new__(cls, conditions_dict, **kwargs):
        """
        Instantiate the appropriate subclass of
        :class:`~pina.data.dataset.PinaDataset`.

        If a graph is present in the conditions, returns a
        :class:`~pina.data.dataset.PinaGraphDataset`, otherwise returns a
        :class:`~pina.data.dataset.PinaTensorDataset`.

        :param dict conditions_dict: Dictionary containing all the conditions
            to be included in the dataset instance.
        :return: A subclass of :class:`~pina.data.dataset.PinaDataset`.
        :rtype: PinaTensorDataset | PinaGraphDataset

        :raises ValueError: If an empty dictionary is provided.
        """

        # Check if conditions_dict is empty
        if len(conditions_dict) == 0:
            raise ValueError("No conditions provided")

        # Check is a Graph is present in the conditions
        is_graph = cls._is_graph_dataset(conditions_dict)
        if is_graph:
            # If a Graph is present, return a PinaGraphDataset
            return PinaGraphDataset(conditions_dict, **kwargs)
        # If no Graph is present, return a PinaTensorDataset
        return PinaTensorDataset(conditions_dict, **kwargs)

    @staticmethod
    def _is_graph_dataset(conditions_dict):
        """
        Check if a graph is present in the conditions (at least one time).

        :param conditions_dict: Dictionary containing the conditions.
        :type conditions_dict: dict
        :return: True if a graph is present in the conditions, False otherwise.
        :rtype: bool
        """

        # Iterate over the conditions dictionary
        for v in conditions_dict.values():
            # Iterate over the values of the current condition
            for cond in v.values():
                # Check if the current value is a list of Data objects
                if isinstance(cond, (Data, Graph, list, tuple)):
                    return True
        return False


class PinaDataset(Dataset, ABC):
    """
    Abstract class for the PINA dataset which extends the PyTorch
    :class:`~torch.utils.data.Dataset` class. It defines the common interface
    for :class:`~pina.data.dataset.PinaTensorDataset` and
    :class:`~pina.data.dataset.PinaGraphDataset` classes.
    """

    def __init__(
        self, conditions_dict, max_conditions_lengths, automatic_batching
    ):
        """
        Initialize the instance by storing the conditions dictionary, the
        maximum number of items per conditions to consider, and the automatic
        batching flag.

        :param dict conditions_dict: A dictionary mapping condition names to
            their respective data. Each key represents a condition name, and the
            corresponding value is a dictionary containing the associated data.
        :param dict max_conditions_lengths: Maximum number of data points that
            can be included in a single batch per condition.
        :param bool automatic_batching: Indicates whether PyTorch automatic
            batching is enabled in
            :class:`~pina.data.data_module.PinaDataModule`.
        """

        # Store the conditions dictionary
        self.conditions_dict = conditions_dict
        # Store the maximum number of conditions to consider
        self.max_conditions_lengths = max_conditions_lengths
        # Store length of each condition
        self.conditions_length = {
            k: len(v["input"]) for k, v in self.conditions_dict.items()
        }
        # Store the maximum length of the dataset
        self.length = max(self.conditions_length.values())
        # Dynamically set the getitem function based on automatic batching
        if automatic_batching:
            self._getitem_func = self._getitem_int
        else:
            self._getitem_func = self._getitem_dummy

    def _get_max_len(self):
        """
        Returns the length of the longest condition in the dataset.

        :return: Length of the longest condition in the dataset.
        :rtype: int
        """

        max_len = 0
        for condition in self.conditions_dict.values():
            max_len = max(max_len, len(condition["input"]))
        return max_len

    def __len__(self):
        return self.length

    def __getitem__(self, idx):
        return self._getitem_func(idx)

    def _getitem_dummy(self, idx):
        """
        Return the index itself. This is used when automatic batching is
        disabled to postpone the data retrieval to the dataloader.

        :param int idx: Index.
        :return: Index.
        :rtype: int
        """

        # If automatic batching is disabled, return the data at the given index
        return idx

    def _getitem_int(self, idx):
        """
        Return the data at the given index in the dataset. This is used when
        automatic batching is enabled.

        :param int idx: Index.
        :return: A dictionary containing the data at the given index.
        :rtype: dict
        """

        # If automatic batching is enabled, return the data at the given index
        return {
            k: {k_data: v[k_data][idx % len(v["input"])] for k_data in v.keys()}
            for k, v in self.conditions_dict.items()
        }

    def get_all_data(self):
        """
        Return all data in the dataset.

        :return: A dictionary containing all the data in the dataset.
        :rtype: dict
        """
        to_return_dict = {}
        for condition, data in self.conditions_dict.items():
            len_condition = len(
                data["input"]
            )  # Length of the current condition
            to_return_dict[condition] = self._retrive_data(
                data, list(range(len_condition))
            )  # Retrieve the data from the current condition
        return to_return_dict

    def fetch_from_idx_list(self, idx):
        """
        Return data from the dataset given a list of indices.

        :param list[int] idx: List of indices.
        :return: A dictionary containing the data at the given indices.
        :rtype: dict
        """

        to_return_dict = {}
        for condition, data in self.conditions_dict.items():
            # Get the indices for the current condition
            cond_idx = idx[: self.max_conditions_lengths[condition]]
            # Get the length of the current condition
            condition_len = self.conditions_length[condition]
            # If the length of the dataset is greater than the length of the
            # current condition, repeat the indices
            if self.length > condition_len:
                cond_idx = [idx % condition_len for idx in cond_idx]
            # Retrieve the data from the current condition
            to_return_dict[condition] = self._retrive_data(data, cond_idx)
        return to_return_dict

    @abstractmethod
    def _retrive_data(self, data, idx_list):
        """
        Abstract method to retrieve data from the dataset given a list of
        indices.
        """


class PinaTensorDataset(PinaDataset):
    """
    Dataset class for the PINA dataset with :class:`torch.Tensor` and
    :class:`~pina.label_tensor.LabelTensor` data.
    """

    # Override _retrive_data method for torch.Tensor data
    def _retrive_data(self, data, idx_list):
        """
        Retrieve data from the dataset given a list of indices.

        :param dict data: Dictionary containing the data
            (only :class:`torch.Tensor` or
            :class:`~pina.label_tensor.LabelTensor`).
        :param list[int] idx_list: indices to retrieve.
        :return: Dictionary containing the data at the given indices.
        :rtype: dict
        """

        return {k: v[idx_list] for k, v in data.items()}

    @property
    def input(self):
        """
        Return the input data for the dataset.

        :return: Dictionary containing the input points.
        :rtype: dict
        """
        return {k: v["input"] for k, v in self.conditions_dict.items()}

    def update_data(self, new_conditions_dict):
        """
        Update the dataset with new data.
        This method is used to update the dataset with new data. It replaces
        the current data with the new data provided in the new_conditions_dict
        parameter.

        :param dict new_conditions_dict: Dictionary containing the new data.
        :return: None
        """
        for condition, data in new_conditions_dict.items():
            if condition in self.conditions_dict:
                self.conditions_dict[condition].update(data)
            else:
                self.conditions_dict[condition] = data


class PinaGraphDataset(PinaDataset):
    """
    Dataset class for the PINA dataset with :class:`~torch_geometric.data.Data`
    and :class:`~pina.graph.Graph` data.
    """

    def _create_graph_batch(self, data):
        """
        Create a LabelBatch object from a list of
        :class:`~torch_geometric.data.Data` objects.

        :param data: List of items to collate in a single batch.
        :type data: list[Data] | list[Graph]
        :return: LabelBatch object all the graph collated in a single batch
            disconnected graphs.
        :rtype: LabelBatch
        """
        batch = LabelBatch.from_data_list(data)
        return batch

    def _create_tensor_batch(self, data):
        """
        Reshape properly ``data`` tensor to be processed handle by the graph
        based models.

        :param data: torch.Tensor object of shape ``(N, ...)`` where ``N`` is
            the number of data objects.
        :type data: torch.Tensor | LabelTensor
        :return: Reshaped tensor object.
        :rtype: torch.Tensor | LabelTensor
        """
        out = data.reshape(-1, *data.shape[2:])
        return out

    def create_batch(self, data):
        """
        Create a Batch object from a list of :class:`~torch_geometric.data.Data`
        objects.

        :param data: List of items to collate in a single batch.
        :type data: list[Data] | list[Graph]
        :return: Batch object.
        :rtype: :class:`~torch_geometric.data.Batch`
            | :class:`~pina.graph.LabelBatch`
        """

        if isinstance(data[0], Data):
            return self._create_graph_batch(data)
        return self._create_tensor_batch(data)

    # Override _retrive_data method for graph handling
    def _retrive_data(self, data, idx_list):
        """
        Retrieve data from the dataset given a list of indices.

        :param dict data: Dictionary containing the data.
        :param list[int] idx_list: List of indices to retrieve.
        :return: Dictionary containing the data at the given indices.
        :rtype: dict
        """

        # Return the data from the current condition
        # If the data is a list of Data objects, create a Batch object
        # If the data is a list of torch.Tensor objects, create a torch.Tensor
        return {
            k: (
                self._create_graph_batch([v[i] for i in idx_list])
                if isinstance(v, list)
                else self._create_tensor_batch(v[idx_list])
            )
            for k, v in data.items()
        }

    @property
    def input(self):
        """
        Return the input data for the dataset.

        :return: Dictionary containing the input points.
        :rtype: dict
        """
        return {k: v["input"] for k, v in self.conditions_dict.items()}



"""Module for the Processing Callbacks."""

import copy
import torch

from lightning.pytorch.callbacks import Callback, TQDMProgressBar
from lightning.pytorch.callbacks.progress.progress_bar import (
    get_standard_metrics,
)
from pina.utils import check_consistency


class MetricTracker(Callback):
    """
    Lightning Callback for Metric Tracking.
    """

    def __init__(self, metrics_to_track=None):
        """
        Tracks specified metrics during training.

        :param metrics_to_track: List of metrics to track.
            Defaults to train loss.
        :type metrics_to_track: list[str], optional
        """
        super().__init__()
        self._collection = []
        # Default to tracking 'train_loss' if not specified
        self.metrics_to_track = metrics_to_track

    def setup(self, trainer, pl_module, stage):
        """
        Called when fit, validate, test, predict, or tune begins.

        :param Trainer trainer: A :class:`~pina.trainer.Trainer` instance.
        :param SolverInterface pl_module: A
            :class:`~pina.solver.solver.SolverInterface` instance.
        :param str stage: Either 'fit', 'test' or 'predict'.
        """
        if self.metrics_to_track is None and trainer.batch_size is None:
            self.metrics_to_track = ["train_loss"]
        elif self.metrics_to_track is None:
            self.metrics_to_track = ["train_loss_epoch"]
        return super().setup(trainer, pl_module, stage)

    def on_train_epoch_end(self, trainer, pl_module):
        """
        Collect and track metrics at the end of each training epoch.

        :param trainer: The trainer object managing the training process.
        :type trainer: pytorch_lightning.Trainer
        :param pl_module: The model being trained (not used here).
        """
        # Track metrics after the first epoch onwards
        if trainer.current_epoch > 0:
            # Append only the tracked metrics to avoid unnecessary data
            tracked_metrics = {
                k: v
                for k, v in trainer.logged_metrics.items()
                if k in self.metrics_to_track
            }
            self._collection.append(copy.deepcopy(tracked_metrics))

    @property
    def metrics(self):
        """
        Aggregate collected metrics over all epochs.

        :return: A dictionary containing aggregated metric values.
        :rtype: dict
        """
        if not self._collection:
            return {}

        # Get intersection of keys across all collected dictionaries
        common_keys = set(self._collection[0]).intersection(
            *self._collection[1:]
        )

        # Stack the metric values for common keys and return
        return {
            k: torch.stack([dic[k] for dic in self._collection])
            for k in common_keys
            if k in self.metrics_to_track
        }


class PINAProgressBar(TQDMProgressBar):
    """
    PINA Implementation of a Lightning Callback for enriching the progress bar.
    """

    BAR_FORMAT = (
        "{l_bar}{bar}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}, "
        "{rate_noinv_fmt}{postfix}]"
    )

    def __init__(self, metrics="val", **kwargs):
        """
        This class enables the display of only relevant metrics during training.

        :param metrics: Logged metrics to be shown during the training.
            Must be a subset of the conditions keys defined in
            :obj:`pina.condition.Condition`.
        :type metrics: str | list(str) | tuple(str)

        :Keyword Arguments:
            The additional keyword arguments specify the progress bar and can be
            choosen from the `pytorch-lightning TQDMProgressBar API
            <https://lightning.ai/docs/pytorch/stable/_modules/lightning/pytorch/callback/progress/tqdm_progress.html#TQDMProgressBar>`_

        Example:
            >>> pbar = PINAProgressBar(['mean'])
            >>> # ... Perform training ...
            >>> trainer = Trainer(solver, callbacks=[pbar])
        """
        super().__init__(**kwargs)
        # check consistency
        if not isinstance(metrics, (list, tuple)):
            metrics = [metrics]
        check_consistency(metrics, str)
        self._sorted_metrics = metrics

    def get_metrics(self, trainer, pl_module):
        r"""Combine progress bar metrics collected from the trainer with
        standard metrics from get_standard_metrics.
        Override this method to customize the items shown in the progress bar.
        The progress bar metrics are sorted according to ``metrics``.

        Here is an example of how to override the defaults:

        .. code-block:: python

            def get_metrics(self, trainer, model):
                # don't show the version number
                items = super().get_metrics(trainer, model)
                items.pop("v_num", None)
                return items

        :return: Dictionary with the items to be displayed in the progress bar.
        :rtype: tuple(dict)
        """
        standard_metrics = get_standard_metrics(trainer)
        pbar_metrics = trainer.progress_bar_metrics
        if pbar_metrics:
            pbar_metrics = {
                key: pbar_metrics[key]
                for key in pbar_metrics
                if key in self._sorted_metrics
            }
        return {**standard_metrics, **pbar_metrics}

    def setup(self, trainer, pl_module, stage):
        """
        Check that the initialized metrics are available and correctly logged.

        :param trainer: The trainer object managing the training process.
        :type trainer: pytorch_lightning.Trainer
        :param pl_module: Placeholder argument.
        """
        # Check if all keys in sort_keys are present in the dictionary
        for key in self._sorted_metrics:
            if (
                key not in trainer.solver.problem.conditions.keys()
                and key != "train"
                and key != "val"
            ):
                raise KeyError(f"Key '{key}' is not present in the dictionary")
        # add the loss pedix
        if trainer.batch_size is not None:
            pedix = "_loss_epoch"
        else:
            pedix = "_loss"
        self._sorted_metrics = [
            metric + pedix for metric in self._sorted_metrics
        ]
        return super().setup(trainer, pl_module, stage)



"""Module for the AbstractProblem class."""

from abc import ABCMeta, abstractmethod
import warnings
from copy import deepcopy
from ..utils import check_consistency
from ..domain import DomainInterface, CartesianDomain
from ..condition.domain_equation_condition import DomainEquationCondition
from ..label_tensor import LabelTensor
from ..utils import merge_tensors, custom_warning_format


class AbstractProblem(metaclass=ABCMeta):
    """
    Abstract base class for PINA problems. All specific problem types should
    inherit from this class.

    A PINA problem is defined by key components, which typically include output
    variables, conditions, and domains over which the conditions are applied.
    """

    def __init__(self):
        """
        Initialization of the :class:`AbstractProblem` class.
        """
        self._discretised_domains = {}

        # create hook conditions <-> problems
        for condition_name in self.conditions:
            self.conditions[condition_name].problem = self

        # Store in domains dict all the domains object directly passed to
        # ConditionInterface. Done for back compatibility with PINA <0.2
        if not hasattr(self, "domains"):
            self.domains = {}
        for cond_name, cond in self.conditions.items():
            if isinstance(cond, DomainEquationCondition):
                if isinstance(cond.domain, DomainInterface):
                    self.domains[cond_name] = cond.domain
                    cond.domain = cond_name

        self._collected_data = {}

    @property
    def collected_data(self):
        """
        Return the collected data from the problem's conditions. If some domains
        are not sampled, they will not be returned by collected data.

        :return: The collected data. Keys are condition names, and values are
            dictionaries containing the input points and the corresponding
            equations or target points.
        :rtype: dict
        """
        # collect data so far
        self.collect_data()
        # raise warning if some sample data are missing
        if not self.are_all_domains_discretised:
            warnings.formatwarning = custom_warning_format
            warnings.filterwarnings("always", category=RuntimeWarning)
            warning_message = "\n".join(
                [
                    f"""{" " * 13} ---> Domain {key} {
                    "sampled" if key in self.discretised_domains 
                    else
                    "not sampled"}"""
                    for key in self.domains
                ]
            )
            warnings.warn(
                "Some of the domains are still not sampled. Consider calling "
                "problem.discretise_domain function for all domains before "
                "accessing the collected data:\n"
                f"{warning_message}",
                RuntimeWarning,
            )
        return self._collected_data

    #  back compatibility 0.1
    @property
    def input_pts(self):
        """
        Return a dictionary mapping condition names to their corresponding
        input points. If some domains are not sampled, they will not be returned
        and the corresponding condition will be empty.

        :return: The input points of the problem.
        :rtype: dict
        """
        to_return = {}
        for cond_name, data in self.collected_data.items():
            to_return[cond_name] = data["input"]
        return to_return

    @property
    def discretised_domains(self):
        """
        Return a dictionary mapping domains to their corresponding sampled
        points.

        :return: The discretised domains.
        :rtype: dict
        """
        return self._discretised_domains

    def __deepcopy__(self, memo):
        """
        Perform a deep copy of the :class:`AbstractProblem` instance.

        :param dict memo: A dictionary used to track objects already copied
            during the deep copy process to prevent redundant copies.
        :return: A deep copy of the :class:`AbstractProblem` instance.
        :rtype: AbstractProblem
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        return result

    @property
    def are_all_domains_discretised(self):
        """
        Check if all the domains are discretised.

        :return: ``True`` if all domains are discretised, ``False`` otherwise.
        :rtype: bool
        """
        return all(
            domain in self.discretised_domains for domain in self.domains
        )

    @property
    def input_variables(self):
        """
        Get the input variables of the problem.

        :return: The input variables of the problem.
        :rtype: list[str]
        """
        variables = []

        if hasattr(self, "spatial_variables"):
            variables += self.spatial_variables
        if hasattr(self, "temporal_variable"):
            variables += self.temporal_variable
        if hasattr(self, "parameters"):
            variables += self.parameters

        return variables

    @input_variables.setter
    def input_variables(self, variables):
        """
        Set the input variables of the AbstractProblem.

        :param list[str] variables: The input variables of the problem.
        :raises RuntimeError: Not implemented.
        """
        raise RuntimeError

    @property
    @abstractmethod
    def output_variables(self):
        """
        Get the output variables of the problem.
        """

    @property
    @abstractmethod
    def conditions(self):
        """
        Get the conditions of the problem.

        :return: The conditions of the problem.
        :rtype: dict
        """
        return self.conditions

    def discretise_domain(
        self, n=None, mode="random", domains="all", sample_rules=None
    ):
        """
        Discretize the problem's domains by sampling a specified number of
        points according to the selected sampling mode.

        :param int n: The number of points to sample.
        :param mode: The sampling method. Default is ``random``.
            Available modes include: random sampling, ``random``;
            latin hypercube sampling, ``latin`` or ``lh``;
            chebyshev sampling, ``chebyshev``; grid sampling ``grid``.
        :param domains: The domains from which to sample. Default is ``all``.
        :type domains: str | list[str]
        :param dict sample_rules: A dictionary defining custom sampling rules
            for input variables. If provided, it must contain a dictionary
            specifying the sampling rule for each variable, overriding the
            ``n`` and ``mode`` arguments. Each key must correspond to the
            input variables from
            :meth:~pina.problem.AbstractProblem.input_variables, and its value
            should be another dictionary with
            two keys: ``n`` (number of points to sample) and ``mode``
            (sampling method). Defaults to None.
        :raises RuntimeError: If both ``n`` and ``sample_rules`` are specified.
        :raises RuntimeError: If neither ``n`` nor ``sample_rules`` are set.

        :Example:
            >>> problem.discretise_domain(n=10, mode='grid')
            >>> problem.discretise_domain(n=10, mode='grid', domains=['gamma1'])
            >>> problem.discretise_domain(
            ...     sample_rules={
            ...         'x': {'n': 10, 'mode': 'grid'},
            ...         'y': {'n': 100, 'mode': 'grid'}
            ...     },
            ...     domains=['D']
            ... )

        .. warning::
            ``random`` is currently the only implemented ``mode`` for all
            geometries, i.e. :class:`~pina.domain.ellipsoid.EllipsoidDomain`,
            :class:`~pina.domain.cartesian.CartesianDomain`,
            :class:`~pina.domain.simplex.SimplexDomain`, and geometry
            compositions :class:`~pina.domain.union_domain.Union`,
            :class:`~pina.domain.difference_domain.Difference`,
            :class:`~pina.domain.exclusion_domain.Exclusion`, and
            :class:`~pina.domain.intersection_domain.Intersection`.
            The modes ``latin`` or ``lh``,  ``chebyshev``, ``grid`` are only
            implemented for :class:`~pina.domain.cartesian.CartesianDomain`.

        .. warning::
            If custom discretisation is applied by setting ``sample_rules`` not
            to ``None``, then the discretised domain must be of class
            :class:`~pina.domain.cartesian.CartesianDomain`
        """

        # check consistecy n, mode, variables, locations
        if sample_rules is not None:
            check_consistency(sample_rules, dict)
        if mode is not None:
            check_consistency(mode, str)
        check_consistency(domains, (list, str))

        # check correct location
        if domains == "all":
            domains = self.domains.keys()
        elif not isinstance(domains, (list)):
            domains = [domains]
        if n is not None and sample_rules is None:
            self._apply_default_discretization(n, mode, domains)
        if n is None and sample_rules is not None:
            self._apply_custom_discretization(sample_rules, domains)
        elif n is not None and sample_rules is not None:
            raise RuntimeError(
                "You can't specify both n and sample_rules at the same time."
            )
        elif n is None and sample_rules is None:
            raise RuntimeError("You have to specify either n or sample_rules.")

    def _apply_default_discretization(self, n, mode, domains):
        """
        Apply default discretization to the problem's domains.

        :param int n: The number of points to sample.
        :param mode: The sampling method.
        :param domains: The domains from which to sample.
        :type domains: str | list[str]
        """
        for domain in domains:
            self.discretised_domains[domain] = (
                self.domains[domain].sample(n, mode).sort_labels()
            )

    def _apply_custom_discretization(self, sample_rules, domains):
        """
        Apply custom discretization to the problem's domains.

        :param dict sample_rules: A dictionary of custom sampling rules.
        :param domains: The domains from which to sample.
        :type domains: str | list[str]
        :raises RuntimeError: If the keys of the sample_rules dictionary are not
            the same as the input variables.
        :raises RuntimeError: If custom discretisation is applied on a domain
            that is not a CartesianDomain.
        """
        if sorted(list(sample_rules.keys())) != sorted(self.input_variables):
            raise RuntimeError(
                "The keys of the sample_rules dictionary must be the same as "
                "the input variables."
            )
        for domain in domains:
            if not isinstance(self.domains[domain], CartesianDomain):
                raise RuntimeError(
                    "Custom discretisation can be applied only on Cartesian "
                    "domains"
                )
            discretised_tensor = []
            for var, rules in sample_rules.items():
                n, mode = rules["n"], rules["mode"]
                points = self.domains[domain].sample(n, mode, var)
                discretised_tensor.append(points)

            self.discretised_domains[domain] = merge_tensors(
                discretised_tensor
            ).sort_labels()

    def add_points(self, new_points_dict):
        """
        Add new points to an already sampled domain.

        :param dict new_points_dict: The dictionary mapping new points to their
            corresponding domain.
        """
        for k, v in new_points_dict.items():
            self.discretised_domains[k] = LabelTensor.vstack(
                [self.discretised_domains[k], v]
            )

    def collect_data(self):
        """
        Aggregate data from the problem's conditions into a single dictionary.
        """
        data = {}
        # Iterate over the conditions and collect data
        for condition_name in self.conditions:
            condition = self.conditions[condition_name]
            # Check if the condition has an domain attribute
            if hasattr(condition, "domain"):
                # Only store the discretisation points if the domain is
                # in the dictionary
                if condition.domain in self.discretised_domains:
                    samples = self.discretised_domains[condition.domain]
                    data[condition_name] = {
                        "input": samples,
                        "equation": condition.equation,
                    }
            else:
                # If the condition does not have a domain attribute, store
                # the input and target points
                keys = condition.__slots__
                values = [getattr(condition, name) for name in keys]
                data[condition_name] = dict(zip(keys, values))
        self._collected_data = data



"""Module for the message passing blocks of the graph neural models."""

__all__ = [
    "InteractionNetworkBlock",
    "DeepTensorNetworkBlock",
    "EnEquivariantNetworkBlock",
    "RadialFieldNetworkBlock",
]

from .interaction_network_block import InteractionNetworkBlock
from .deep_tensor_network_block import DeepTensorNetworkBlock
from .en_equivariant_network_block import EnEquivariantNetworkBlock
from .radial_field_network_block import RadialFieldNetworkBlock



"""Module for the PowerLoss class."""

import torch

from ..utils import check_consistency
from .loss_interface import LossInterface


class PowerLoss(LossInterface):
    r"""
    Implementation of the Power Loss. It defines a criterion to measures the 
    pointwise error between values in the input :math:`x` and values in the
    target :math:`y`.

    If ``reduction`` is set to ``none``, the loss can be written as:

    .. math::
        \ell(x, y) = L = \{l_1,\dots,l_N\}^\top, \quad
        l_n = \frac{1}{D}\left[\sum_{i=1}^{D} 
        \left| x_n^i - y_n^i \right|^p\right],
    
    If ``relative`` is set to ``True``, the relative error is computed:

    .. math::
        \ell(x, y) = L = \{l_1,\dots,l_N\}^\top, \quad
        l_n = \frac{ \sum_{i=1}^{D} | x_n^i - y_n^i|^p }
        {\sum_{i=1}^{D}|y_n^i|^p},

    where :math:`N` is the batch size.
    
    If ``reduction`` is not ``none``, then:

    .. math::
        \ell(x, y) =
        \begin{cases}
            \operatorname{mean}(L), & \text{if reduction} = \text{`mean';}\\
            \operatorname{sum}(L),  & \text{if reduction} = \text{`sum'.}
        \end{cases}
    """

    def __init__(self, p=2, reduction="mean", relative=False):
        """
        Initialization of the :class:`PowerLoss` class.

        :param int p: Degree of the Lp norm. It specifies the norm to be
            computed. Default is ``2`` (euclidean norm).
        :param str reduction: The reduction method for the loss.
            Available options: ``none``, ``mean``, ``sum``.
            If ``none``, no reduction is applied. If ``mean``, the sum of the
            loss values is divided by the number of values. If ``sum``, the loss
            values are summed. Default is ``mean``.
        :param bool relative: If ``True``, the relative error is computed.
            Default is ``False``.
        """
        super().__init__(reduction=reduction)

        # check consistency
        check_consistency(p, (str, int, float))
        check_consistency(relative, bool)

        self.p = p
        self.relative = relative

    def forward(self, input, target):
        """
        Forward method of the loss function.

        :param torch.Tensor input: Input tensor from real data.
        :param torch.Tensor target: Model tensor output.
        :return: Loss evaluation.
        :rtype: torch.Tensor
        """
        loss = torch.abs((input - target)).pow(self.p).mean(-1)
        if self.relative:
            loss = loss / torch.abs(input).pow(self.p).mean(-1)
        return self._reduction(loss)



"""Module for the Scalar Weighting."""

from .weighting_interface import WeightingInterface
from ..utils import check_consistency


class _NoWeighting(WeightingInterface):
    """
    Weighting scheme that does not apply any weighting to the losses.
    """

    def aggregate(self, losses):
        """
        Aggregate the losses.

        :param dict losses: The dictionary of losses.
        :return: The aggregated losses.
        :rtype: torch.Tensor
        """
        return sum(losses.values())


class ScalarWeighting(WeightingInterface):
    """
    Weighting scheme that assigns a scalar weight to each loss term.
    """

    def __init__(self, weights):
        """
        Initialization of the :class:`ScalarWeighting` class.

        :param weights: The weights to be assigned to each loss term.
            If a single scalar value is provided, it is assigned to all loss
            terms. If a dictionary is provided, the keys are the conditions and
            the values are the weights. If a condition is not present in the
            dictionary, the default value is used.
        :type weights: float | int | dict
        """
        super().__init__()
        check_consistency([weights], (float, dict, int))
        if isinstance(weights, (float, int)):
            self.default_value_weights = weights
            self.weights = {}
        else:
            self.default_value_weights = 1
            self.weights = weights

    def aggregate(self, losses):
        """
        Aggregate the losses.

        :param dict losses: The dictionary of losses.
        :return: The aggregated losses.
        :rtype: torch.Tensor
        """
        return sum(
            self.weights.get(condition, self.default_value_weights) * loss
            for condition, loss in losses.items()
        )



"""Module for the Condition interface."""

from abc import ABCMeta
from torch_geometric.data import Data
from ..label_tensor import LabelTensor
from ..graph import Graph


class ConditionInterface(metaclass=ABCMeta):
    """
    Abstract class which defines a common interface for all the conditions.
    It defined a common interface for all the conditions.

    """

    def __init__(self):
        """
        Initialize the ConditionInterface object.
        """

        self._problem = None

    @property
    def problem(self):
        """
        Return the problem to which the condition is associated.

        :return: Problem to which the condition is associated.
        :rtype: ~pina.problem.abstract_problem.AbstractProblem
        """
        return self._problem

    @problem.setter
    def problem(self, value):
        """
        Set the problem to which the condition is associated.

        :param pina.problem.abstract_problem.AbstractProblem value: Problem to
            which the condition is associated
        """
        self._problem = value

    @staticmethod
    def _check_graph_list_consistency(data_list):
        """
        Check the consistency of the list of Data/Graph objects. It performs
        the following checks:

        1. All elements in the list must be of the same type (either Data or
        Graph).
        2. All elements in the list must have the same keys.
        3. The type of each tensor must be consistent across all elements in
        the list.
        4. If the tensor is a LabelTensor, the labels must be consistent across
        all elements in the list.

        :param data_list: List of Data/Graph objects to check
        :type data_list: list[Data] | list[Graph] | tuple[Data] | tuple[Graph]

        :raises ValueError:  If the input types are invalid.
        :raises ValueError: If all elements in the list do not have the same
            keys.
        :raises ValueError: If the type of each tensor is not consistent across
            all elements in the list.
        :raises ValueError: If the labels of the LabelTensors are not consistent
            across all elements in the list.
        """

        # If the data is a Graph or Data object, return (do not need to check
        # anything)
        if isinstance(data_list, (Graph, Data)):
            return

        # check all elements in the list are of the same type
        if not all(isinstance(i, (Graph, Data)) for i in data_list):
            raise ValueError(
                "Invalid input types. "
                "Please provide either Data or Graph objects."
            )
        data = data_list[0]
        # Store the keys of the first element in the list
        keys = sorted(list(data.keys()))

        # Store the type of each tensor inside first element Data/Graph object
        data_types = {name: tensor.__class__ for name, tensor in data.items()}

        # Store the labels of each LabelTensor inside first element Data/Graph
        # object
        labels = {
            name: tensor.labels
            for name, tensor in data.items()
            if isinstance(tensor, LabelTensor)
        }

        # Iterate over the list of Data/Graph objects
        for data in data_list[1:]:
            # Check if the keys of the current element are the same as the first
            # element
            if sorted(list(data.keys())) != keys:
                raise ValueError(
                    "All elements in the list must have the same keys."
                )
            for name, tensor in data.items():
                # Check if the type of each tensor inside the current element
                # is the same as the first element
                if tensor.__class__ is not data_types[name]:
                    raise ValueError(
                        f"Data {name} must be a {data_types[name]}, got "
                        f"{tensor.__class__}"
                    )
                # If the tensor is a LabelTensor, check if the labels are the
                # same as the first element
                if isinstance(tensor, LabelTensor):
                    if tensor.labels != labels[name]:
                        raise ValueError(
                            "LabelTensor must have the same labels"
                        )

    def __getattribute__(self, name):
        to_return = super().__getattribute__(name)
        if isinstance(to_return, (Graph, Data)):
            to_return = [to_return]
        return to_return



"""Old module for geometry classes and functions. Deprecated in 0.2.0."""

import warnings

from ..domain import *
from ..utils import custom_warning_format

# back-compatibility 0.1
# creating alias
Location = DomainInterface

# Set the custom format for warnings
warnings.formatwarning = custom_warning_format
warnings.filterwarnings("always", category=DeprecationWarning)
warnings.warn(
    "'pina.geometry' is deprecated and will be removed "
    "in future versions. Please use 'pina.domain' instead. "
    "Location moved to DomainInferface object.",
    DeprecationWarning,
)



"""Module for the SpatialProblem class."""

from abc import abstractmethod

from .abstract_problem import AbstractProblem


class SpatialProblem(AbstractProblem):
    """
    Class for defining spatial problems, where the problem domain is defined in
    terms of spatial variables.
    """

    @abstractmethod
    def spatial_domain(self):
        """
        The spatial domain of the problem.
        """

    @property
    def spatial_variables(self):
        """
        Get the spatial input variables of the problem.

        :return: The spatial input variables of the problem.
        :rtype: list[str]
        """
        return self.spatial_domain.variables



"""Module for the Optimizers and Schedulers."""

__all__ = [
    "Optimizer",
    "TorchOptimizer",
    "Scheduler",
    "TorchScheduler",
]

from .optimizer_interface import Optimizer
from .torch_optimizer import TorchOptimizer
from .scheduler_interface import Scheduler
from .torch_scheduler import TorchScheduler



"""Module for the LinearWeightUpdate callback."""

import warnings
from lightning.pytorch.callbacks import Callback
from ..utils import check_consistency
from ..loss import ScalarWeighting


class LinearWeightUpdate(Callback):
    """
    Callback to linearly adjust the weight of a condition from an
    initial value to a target value over a specified number of epochs.
    """

    def __init__(
        self, target_epoch, condition_name, initial_value, target_value
    ):
        """
        Callback initialization.

        :param int target_epoch: The epoch at which the weight of the condition
            should reach the target value.
        :param str condition_name: The name of the condition whose weight
            should be adjusted.
        :param float initial_value: The initial value of the weight.
        :param float target_value: The target value of the weight.
        """
        super().__init__()
        self.target_epoch = target_epoch
        self.condition_name = condition_name
        self.initial_value = initial_value
        self.target_value = target_value

        # Check consistency
        check_consistency(self.target_epoch, int, subclass=False)
        check_consistency(self.condition_name, str, subclass=False)
        check_consistency(self.initial_value, (float, int), subclass=False)
        check_consistency(self.target_value, (float, int), subclass=False)

    def on_train_start(self, trainer, pl_module):
        """
        Initialize the weight of the condition to the specified `initial_value`.

        :param Trainer trainer: A :class:`~pina.trainer.Trainer` instance.
        :param SolverInterface pl_module: A
            :class:`~pina.solver.solver.SolverInterface` instance.
        """
        # Check that the target epoch is valid
        if not 0 < self.target_epoch <= trainer.max_epochs:
            raise ValueError(
                "`target_epoch` must be greater than 0"
                " and less than or equal to `max_epochs`."
            )

        # Check that the condition is a problem condition
        if self.condition_name not in pl_module.problem.conditions:
            raise ValueError(
                f"`{self.condition_name}` must be a problem condition."
            )

        # Check that the initial value is not equal to the target value
        if self.initial_value == self.target_value:
            warnings.warn(
                "`initial_value` is equal to `target_value`. "
                "No effective adjustment will be performed.",
                UserWarning,
            )

        # Check that the weighting schema is ScalarWeighting
        if not isinstance(pl_module.weighting, ScalarWeighting):
            raise ValueError("The weighting schema must be ScalarWeighting.")

        # Initialize the weight of the condition
        pl_module.weighting.weights[self.condition_name] = self.initial_value

    def on_train_epoch_start(self, trainer, pl_module):
        """
        Adjust at each epoch the weight of the condition.

        :param Trainer trainer: A :class:`~pina.trainer.Trainer` instance.
        :param SolverInterface pl_module: A
            :class:`~pina.solver.solver.SolverInterface` instance.
        """
        if 0 < trainer.current_epoch <= self.target_epoch:
            pl_module.weighting.weights[self.condition_name] += (
                self.target_value - self.initial_value
            ) / (self.target_epoch - 1)



"""Module for the Residual-Based Attention PINN solver."""

from copy import deepcopy
import torch

from .pinn import PINN
from ...utils import check_consistency


class RBAPINN(PINN):
    r"""
    Residual-based Attention Physics-Informed Neural Network (RBAPINN) solver
    class. This class implements the Residual-based Attention Physics-Informed
    Neural Network solver, using a user specified ``model`` to solve a specific
    ``problem``. It can be used to solve both forward and inverse problems.

    The Residual-based Attention Physics-Informed Neural Network solver aims to
    find the solution :math:`\mathbf{u}:\Omega\rightarrow\mathbb{R}^m` of a
    differential problem:

    .. math::

        \begin{cases}
        \mathcal{A}[\mathbf{u}](\mathbf{x})=0\quad,\mathbf{x}\in\Omega\\
        \mathcal{B}[\mathbf{u}](\mathbf{x})=0\quad,
        \mathbf{x}\in\partial\Omega
        \end{cases}
    
    minimizing the loss function:

    .. math::

        \mathcal{L}_{\rm{problem}} = \frac{1}{N} \sum_{i=1}^{N_\Omega} 
        \lambda_{\Omega}^{i} \mathcal{L} \left( \mathcal{A}
        [\mathbf{u}](\mathbf{x}) \right) + \frac{1}{N} 
        \sum_{i=1}^{N_{\partial\Omega}}
        \lambda_{\partial\Omega}^{i} \mathcal{L} 
        \left( \mathcal{B}[\mathbf{u}](\mathbf{x})
        \right),
    
    denoting the weights as:
    :math:`\lambda_{\Omega}^1, \dots, \lambda_{\Omega}^{N_\Omega}` and
    :math:`\lambda_{\partial \Omega}^1, \dots, 
    \lambda_{\Omega}^{N_\partial \Omega}`
    for :math:`\Omega` and :math:`\partial \Omega`, respectively.

    Residual-based Attention Physics-Informed Neural Network updates the weights
    of the residuals at every epoch as follows:

    .. math::

        \lambda_i^{k+1} \leftarrow \gamma\lambda_i^{k} + 
        \eta\frac{\lvert r_i\rvert}{\max_j \lvert r_j\rvert},

    where :math:`r_i` denotes the residual at point :math:`i`, :math:`\gamma`
    denotes the decay rate, and :math:`\eta` is the learning rate for the
    weights' update.

    .. seealso::
        **Original reference**: Sokratis J. Anagnostopoulos, Juan D. Toscano,
        Nikolaos Stergiopulos, and George E. Karniadakis.
        *Residual-based attention and connection to information 
        bottleneck theory in PINNs.*
        Computer Methods in Applied Mechanics and Engineering 421 (2024): 116805
        DOI: `10.1016/j.cma.2024.116805
        <https://doi.org/10.1016/j.cma.2024.116805>`_.
    """

    def __init__(
        self,
        problem,
        model,
        optimizer=None,
        scheduler=None,
        weighting=None,
        loss=None,
        eta=0.001,
        gamma=0.999,
    ):
        """
        Initialization of the :class:`RBAPINN` class.

        :param AbstractProblem problem: The problem to be solved.
        :param torch.nn.Module model: The neural network model to be used.
        :param Optimizer optimizer: The optimizer to be used.
            If ``None``, the :class:`torch.optim.Adam` optimizer is used.
            Default is ``None``.
        :param Scheduler scheduler: Learning rate scheduler.
            If ``None``, the :class:`torch.optim.lr_scheduler.ConstantLR`
            scheduler is used. Default is ``None``.
        :param WeightingInterface weighting: The weighting schema to be used.
            If ``None``, no weighting schema is used. Default is ``None``.
        :param torch.nn.Module loss: The loss function to be minimized.
            If ``None``, the :class:`torch.nn.MSELoss` loss is used.
            Default is `None`.
        :param float | int eta: The learning rate for the weights of the
            residuals. Default is ``0.001``.
        :param float gamma: The decay parameter in the update of the weights
            of the residuals. Must be between ``0`` and ``1``.
            Default is ``0.999``.
        """
        super().__init__(
            model=model,
            problem=problem,
            optimizer=optimizer,
            scheduler=scheduler,
            weighting=weighting,
            loss=loss,
        )

        # check consistency
        check_consistency(eta, (float, int))
        check_consistency(gamma, float)
        assert (
            0 < gamma < 1
        ), f"Invalid range: expected 0 < gamma < 1, got {gamma=}"
        self.eta = eta
        self.gamma = gamma

        # initialize weights
        self.weights = {}
        for condition_name in problem.conditions:
            self.weights[condition_name] = 0

        # define vectorial loss
        self._vectorial_loss = deepcopy(self.loss)
        self._vectorial_loss.reduction = "none"

    # for now RBAPINN is implemented only for batch_size = None
    def on_train_start(self):
        """
        Hook method called at the beginning of training.

        :raises NotImplementedError: If the batch size is not ``None``.
        """
        if self.trainer.batch_size is not None:
            raise NotImplementedError(
                "RBAPINN only works with full batch "
                "size, set batch_size=None inside the "
                "Trainer to use the solver."
            )
        return super().on_train_start()

    def _vect_to_scalar(self, loss_value):
        """
        Computation of the scalar loss.

        :param LabelTensor loss_value: the tensor of pointwise losses.
        :raises RuntimeError: If the loss reduction is not ``mean`` or ``sum``.
        :return: The computed scalar loss.
        :rtype: LabelTensor
        """
        if self.loss.reduction == "mean":
            ret = torch.mean(loss_value)
        elif self.loss.reduction == "sum":
            ret = torch.sum(loss_value)
        else:
            raise RuntimeError(
                f"Invalid reduction, got {self.loss.reduction} "
                "but expected mean or sum."
            )
        return ret

    def loss_phys(self, samples, equation):
        """
        Computes the physics loss for the physics-informed solver based on the
        provided samples and equation.

        :param LabelTensor samples: The samples to evaluate the physics loss.
        :param EquationInterface equation: The governing equation.
        :return: The computed physics loss.
        :rtype: LabelTensor
        """
        residual = self.compute_residual(samples=samples, equation=equation)
        cond = self.current_condition_name

        r_norm = (
            self.eta
            * torch.abs(residual)
            / (torch.max(torch.abs(residual)) + 1e-12)
        )
        self.weights[cond] = (self.gamma * self.weights[cond] + r_norm).detach()

        loss_value = self._vectorial_loss(
            torch.zeros_like(residual, requires_grad=True), residual
        )

        return self._vect_to_scalar(self.weights[cond] ** 2 * loss_value)



"""Module for the PINA Torch Optimizer"""

try:
    from torch.optim.lr_scheduler import LRScheduler  # torch >= 2.0
except ImportError:
    from torch.optim.lr_scheduler import (
        _LRScheduler as LRScheduler,
    )  # torch < 2.0

from ..utils import check_consistency
from .optimizer_interface import Optimizer
from .scheduler_interface import Scheduler


class TorchScheduler(Scheduler):
    """
    A wrapper class for using PyTorch schedulers.
    """

    def __init__(self, scheduler_class, **kwargs):
        """
        Initialization of the :class:`TorchScheduler` class.

        :param torch.optim.LRScheduler scheduler_class: A
            :class:`torch.optim.LRScheduler` class.
        :param dict kwargs: Additional parameters passed to ``scheduler_class``,
            see more
            `here <https://pytorch.org/docs/stable/optim.html#algorithms>_`.
        """
        check_consistency(scheduler_class, LRScheduler, subclass=True)

        self.scheduler_class = scheduler_class
        self.kwargs = kwargs
        self._scheduler_instance = None

    def hook(self, optimizer):
        """
        Initialize the scheduler instance with the given parameters.

        :param dict parameters: The parameters of the optimizer.
        """
        check_consistency(optimizer, Optimizer)
        self._scheduler_instance = self.scheduler_class(
            optimizer.instance, **self.kwargs
        )

    @property
    def instance(self):
        """
        Get the scheduler instance.

        :return: The scheduelr instance.
        :rtype: torch.optim.LRScheduler
        """
        return self._scheduler_instance



"""Module for utility functions for the convolutional layer."""

import torch


def check_point(x, current_stride, dim):
    """
    Check if the point is in the current stride.

    :param torch.Tensor x: The input data.
    :param int current_stride: The current stride.
    :param int dim: The shape of the filter.
    :return: The indeces of the points in the current stride.
    :rtype: torch.Tensor
    """
    max_stride = current_stride + dim
    indeces = torch.logical_and(
        x[..., :-1] < max_stride, x[..., :-1] >= current_stride
    ).all(dim=-1)
    return indeces


def map_points_(x, filter_position):
    """
    The mapping function for n-dimensional case.

    :param torch.Tensor x: The two-dimensional input data.
    :param list[int] filter_position: The position of the filter.
    :return: The data mapped in-place.
    :rtype: torch.tensor
    """
    x.add_(-filter_position)

    return x


def optimizing(f):
    """
    Decorator to call the function only once.

    :param f: python function
    :type f: Callable
    """

    def wrapper(*args, **kwargs):
        """
        Wrapper function.

        :param args: The arguments of the function.
        :param kwargs: The keyword arguments of the function.
        """
        if kwargs["type_"] == "forward":
            if not wrapper.has_run_inverse:
                wrapper.has_run_inverse = True
                return f(*args, **kwargs)

        if kwargs["type_"] == "inverse":
            if not wrapper.has_run:
                wrapper.has_run = True
                return f(*args, **kwargs)

        return f(*args, **kwargs)

    wrapper.has_run_inverse = False
    wrapper.has_run = False

    return wrapper



"""Module for the DataCondition class."""

import torch
from torch_geometric.data import Data
from .condition_interface import ConditionInterface
from ..label_tensor import LabelTensor
from ..graph import Graph


class DataCondition(ConditionInterface):
    """
    Condition defined by input data and conditional variables. It can be used
    in unsupervised learning problems. Based on the type of the input,
    different condition implementations are available:

    - :class:`TensorDataCondition`: For :class:`torch.Tensor` or
      :class:`~pina.label_tensor.LabelTensor` input data.
    - :class:`GraphDataCondition`: For :class:`~pina.graph.Graph` or
      :class:`~torch_geometric.data.Data` input data.
    """

    __slots__ = ["input", "conditional_variables"]
    _avail_input_cls = (torch.Tensor, LabelTensor, Data, Graph, list, tuple)
    _avail_conditional_variables_cls = (torch.Tensor, LabelTensor)

    def __new__(cls, input, conditional_variables=None):
        """
        Instantiate the appropriate subclass of :class:`DataCondition` based on
        the type of ``input``.

        :param input: Input data for the condition.
        :type input: torch.Tensor | LabelTensor | Graph |
            Data | list[Graph] | list[Data] | tuple[Graph] | tuple[Data]
        :param conditional_variables: Conditional variables for the condition.
        :type conditional_variables: torch.Tensor | LabelTensor, optional
        :return: Subclass of DataCondition.
        :rtype: pina.condition.data_condition.TensorDataCondition |
            pina.condition.data_condition.GraphDataCondition

        :raises ValueError: If input is not of type :class:`torch.Tensor`,
            :class:`~pina.label_tensor.LabelTensor`, :class:`~pina.graph.Graph`,
            or :class:`~torch_geometric.data.Data`.
        """

        if cls != DataCondition:
            return super().__new__(cls)
        if isinstance(input, (torch.Tensor, LabelTensor)):
            subclass = TensorDataCondition
            return subclass.__new__(subclass, input, conditional_variables)

        if isinstance(input, (Graph, Data, list, tuple)):
            cls._check_graph_list_consistency(input)
            subclass = GraphDataCondition
            return subclass.__new__(subclass, input, conditional_variables)

        raise ValueError(
            "Invalid input types. "
            "Please provide either torch_geometric.data.Data or Graph objects."
        )

    def __init__(self, input, conditional_variables=None):
        """
        Initialize the object by storing the input and conditional
        variables (if any).

        :param input: Input data for the condition.
        :type input: torch.Tensor | LabelTensor | Graph | Data | list[Graph] |
            list[Data] | tuple[Graph] | tuple[Data]
        :param conditional_variables: Conditional variables for the condition.
        :type conditional_variables: torch.Tensor | LabelTensor

        .. note::
            If ``input`` consists of a list of :class:`~pina.graph.Graph` or
            :class:`~torch_geometric.data.Data`, all elements must have the same
            structure (keys and data types)
        """

        super().__init__()
        self.input = input
        self.conditional_variables = conditional_variables


class TensorDataCondition(DataCondition):
    """
    DataCondition for :class:`torch.Tensor` or
    :class:`~pina.label_tensor.LabelTensor` input data
    """


class GraphDataCondition(DataCondition):
    """
    DataCondition for :class:`~pina.graph.Graph` or
    :class:`~torch_geometric.data.Data` input data
    """



"""Module for the Averaging Neural Operator model class."""

import torch
from torch import nn
from .block.average_neural_operator_block import AVNOBlock
from .kernel_neural_operator import KernelNeuralOperator
from ..utils import check_consistency


class AveragingNeuralOperator(KernelNeuralOperator):
    """
    Averaging Neural Operator model class.

    The Averaging Neural Operator is a general architecture for learning
    operators, which map functions to functions. It can be trained both with
    Supervised and Physics-Informed learning strategies. The Averaging Neural
    Operator performs convolution by means of a field average.

    .. seealso::

        **Original reference**: Lanthaler S., Li, Z., Stuart, A. (2020).
        *The Nonlocal Neural Operator: Universal Approximation*.
        DOI: `arXiv preprint arXiv:2304.13221.
        <https://arxiv.org/abs/2304.13221>`_
    """

    def __init__(
        self,
        lifting_net,
        projecting_net,
        field_indices,
        coordinates_indices,
        n_layers=4,
        func=nn.GELU,
    ):
        """
        Initialization of the :class:`AveragingNeuralOperator` class.

        :param torch.nn.Module lifting_net: The lifting neural network mapping
            the input to its hidden dimension. It must take as input the input
            field and the coordinates at which the input field is evaluated.
        :param torch.nn.Module projecting_net: The projection neural network
            mapping the hidden representation to the output function. It must
            take as input the embedding dimension plus the dimension of the
            coordinates.
        :param list[str] field_indices: The labels of the fields in the input
            tensor.
        :param list[str] coordinates_indices: The labels of the coordinates in
            the input tensor.
        :param int n_layers: The number of hidden layers. Default is ``4``.
        :param torch.nn.Module func: The activation function to use.
            Default is :class:`torch.nn.GELU`.
        :raises ValueError: If the input dimension does not match with the
            labels of the fields and coordinates.
        :raises ValueError: If the input dimension of the projecting network
            does not match with the hidden dimension of the lifting network.
        """

        # check consistency
        check_consistency(field_indices, str)
        check_consistency(coordinates_indices, str)
        check_consistency(n_layers, int)
        check_consistency(func, nn.Module, subclass=True)

        # check hidden dimensions match
        input_lifting_net = next(lifting_net.parameters()).size()[-1]
        output_lifting_net = lifting_net(
            torch.rand(size=next(lifting_net.parameters()).size())
        ).shape[-1]
        projecting_net_input = next(projecting_net.parameters()).size()[-1]

        if len(field_indices) + len(coordinates_indices) != input_lifting_net:
            raise ValueError(
                "The lifting_net must take as input the "
                "coordinates vector and the field vector."
            )

        if (
            output_lifting_net + len(coordinates_indices)
            != projecting_net_input
        ):
            raise ValueError(
                "The projecting_net input must be equal to"
                "the embedding dimension (which is the output) "
                "of the lifting_net plus the dimension of the "
                "coordinates, i.e. len(coordinates_indices)."
            )

        # assign
        self.coordinates_indices = coordinates_indices
        self.field_indices = field_indices
        integral_net = nn.Sequential(
            *[AVNOBlock(output_lifting_net, func) for _ in range(n_layers)]
        )
        super().__init__(lifting_net, integral_net, projecting_net)

    def forward(self, x):
        r"""
        Forward pass for the :class:`AveragingNeuralOperator` model.

        The ``lifting_net`` maps the input to the hidden dimension.
        Then, several layers of
        :class:`~pina.model.block.average_neural_operator_block.AVNOBlock` are
        applied. Finally, the ``projection_net`` maps the hidden representation
        to the output function.

        :param LabelTensor x: The input tensor for performing the computation.
            It expects a tensor :math:`B \times N \times D`, where :math:`B` is
            the batch_size, :math:`N` the number of points in the mesh,
            :math:`D` the dimension of the problem, i.e. the sum
            of ``len(coordinates_indices)`` and ``len(field_indices)``.
        :return: The output tensor.
        :rtype: torch.Tensor
        """
        points_tmp = x.extract(self.coordinates_indices)
        new_batch = x.extract(self.field_indices)
        new_batch = torch.cat((new_batch, points_tmp), dim=-1)
        new_batch = self._lifting_operator(new_batch)
        new_batch = self._integral_kernels(new_batch)
        new_batch = torch.cat((new_batch, points_tmp), dim=-1)
        new_batch = self._projection_operator(new_batch)
        return new_batch



"""Module for the Deep Tensor Network block."""

import torch
from torch_geometric.nn import MessagePassing
from ....utils import check_positive_integer


class DeepTensorNetworkBlock(MessagePassing):
    """
    Implementation of the Deep Tensor Network block.

    This block is used to perform message-passing between nodes and edges in a
    graph neural network, following the scheme proposed by Schutt et al. in
    2017. It serves as an inner block in a larger graph neural network
    architecture.

    The message between two nodes connected by an edge is computed by applying a
    linear transformation to the sender node features and the edge features,
    followed by a non-linear activation function. Messages are then aggregated
    using an aggregation scheme (e.g., sum, mean, min, max, or product).

    The update step is performed by a simple addition of the incoming messages
    to the node features.

    .. seealso::

        **Original reference**: Schutt, K., Arbabzadah, F., Chmiela, S. et al.
        (2017). *Quantum-Chemical Insights from Deep Tensor Neural Networks*.
        Nature Communications 8, 13890 (2017).
        DOI: `<https://doi.org/10.1038/ncomms13890>`_.
    """

    def __init__(
        self,
        node_feature_dim,
        edge_feature_dim,
        activation=torch.nn.Tanh,
        aggr="add",
        node_dim=-2,
        flow="source_to_target",
    ):
        """
        Initialization of the :class:`DeepTensorNetworkBlock` class.

        :param int node_feature_dim: The dimension of the node features.
        :param int edge_feature_dim: The dimension of the edge features.
        :param torch.nn.Module activation: The activation function.
            Default is :class:`torch.nn.Tanh`.
        :param str aggr: The aggregation scheme to use for message passing.
            Available options are "add", "mean", "min", "max", "mul".
            See :class:`torch_geometric.nn.MessagePassing` for more details.
            Default is "add".
        :param int node_dim: The axis along which to propagate. Default is -2.
        :param str flow: The direction of message passing. Available options
            are "source_to_target" and "target_to_source".
            The "source_to_target" flow means that messages are sent from
            the source node to the target node, while the "target_to_source"
            flow means that messages are sent from the target node to the
            source node. See :class:`torch_geometric.nn.MessagePassing` for more
            details. Default is "source_to_target".
        :raises AssertionError: If `node_feature_dim` is not a positive integer.
        :raises AssertionError: If `edge_feature_dim` is not a positive integer.
        """
        super().__init__(aggr=aggr, node_dim=node_dim, flow=flow)

        # Check values
        check_positive_integer(node_feature_dim, strict=True)
        check_positive_integer(edge_feature_dim, strict=True)

        # Activation function
        self.activation = activation()

        # Layer for processing node features
        self.node_layer = torch.nn.Linear(
            in_features=node_feature_dim,
            out_features=node_feature_dim,
            bias=True,
        )

        # Layer for processing edge features
        self.edge_layer = torch.nn.Linear(
            in_features=edge_feature_dim,
            out_features=node_feature_dim,
            bias=True,
        )

        # Layer for computing the message
        self.message_layer = torch.nn.Linear(
            in_features=node_feature_dim,
            out_features=node_feature_dim,
            bias=False,
        )

    def forward(self, x, edge_index, edge_attr):
        """
        Forward pass of the block, triggering the message-passing routine.

        :param x: The node features.
        :type x: torch.Tensor | LabelTensor
        :param torch.Tensor edge_index: The edge indeces.
        :param edge_attr: The edge attributes.
        :type edge_attr: torch.Tensor | LabelTensor
        :return: The updated node features.
        :rtype: torch.Tensor
        """
        return self.propagate(edge_index=edge_index, x=x, edge_attr=edge_attr)

    def message(self, x_j, edge_attr):
        """
        Compute the message to be passed between nodes and edges.

        :param x_j: The node features of the sender nodes.
        :type x_j: torch.Tensor | LabelTensor
        :param edge_attr: The edge attributes.
        :type edge_attr: torch.Tensor | LabelTensor
        :return: The message to be passed.
        :rtype: torch.Tensor
        """
        # Process node and edge features
        filter_node = self.node_layer(x_j)
        filter_edge = self.edge_layer(edge_attr)

        # Compute the message to be passed
        message = self.message_layer(filter_node * filter_edge)

        return self.activation(message)

    def update(self, message, x):
        """
        Update the node features with the received messages.

        :param torch.Tensor message: The message to be passed.
        :param x: The node features.
        :type x: torch.Tensor | LabelTensor
        :return: The updated node features.
        :rtype: torch.Tensor
        """
        return x + message



import pytest
import torch
from pina import LabelTensor
from pina.graph import RadiusGraph, KNNGraph, Graph
from torch_geometric.data import Data


def build_edge_attr(pos, edge_index):
    return torch.cat([pos[edge_index[0]], pos[edge_index[1]]], dim=-1)


@pytest.mark.parametrize(
    "x, pos",
    [
        (torch.rand(10, 2), torch.rand(10, 3)),
        (
            LabelTensor(torch.rand(10, 2), ["u", "v"]),
            LabelTensor(torch.rand(10, 3), ["x", "y", "z"]),
        ),
    ],
)
def test_build_graph(x, pos):
    edge_index = torch.tensor(
        [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [1, 2, 3, 4, 5, 6, 7, 8, 9, 0]],
        dtype=torch.int64,
    )
    graph = Graph(x=x, pos=pos, edge_index=edge_index)
    assert hasattr(graph, "x")
    assert hasattr(graph, "pos")
    assert hasattr(graph, "edge_index")
    assert torch.isclose(graph.x, x).all()
    if isinstance(x, LabelTensor):
        assert isinstance(graph.x, LabelTensor)
        assert graph.x.labels == x.labels
    else:
        assert isinstance(graph.pos, torch.Tensor)
    assert torch.isclose(graph.pos, pos).all()
    if isinstance(pos, LabelTensor):
        assert isinstance(graph.pos, LabelTensor)
        assert graph.pos.labels == pos.labels
    else:
        assert isinstance(graph.pos, torch.Tensor)

    edge_index = torch.tensor(
        [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [1, 2, 3, 4, 5, 6, 7, 8, 9, 0]],
        dtype=torch.int64,
    )
    graph = Graph(x=x, edge_index=edge_index)
    assert hasattr(graph, "x")
    assert hasattr(graph, "pos")
    assert hasattr(graph, "edge_index")
    assert torch.isclose(graph.x, x).all()
    if isinstance(x, LabelTensor):
        assert isinstance(graph.x, LabelTensor)
        assert graph.x.labels == x.labels
    else:
        assert isinstance(graph.x, torch.Tensor)


@pytest.mark.parametrize(
    "x, pos",
    [
        (torch.rand(10, 2), torch.rand(10, 3)),
        (
            LabelTensor(torch.rand(10, 2), ["u", "v"]),
            LabelTensor(torch.rand(10, 3), ["x", "y", "z"]),
        ),
    ],
)
@pytest.mark.parametrize("loop", [True, False])
def test_build_radius_graph(x, pos, loop):
    graph = RadiusGraph(x=x, pos=pos, radius=0.5, loop=loop)
    assert hasattr(graph, "x")
    assert hasattr(graph, "pos")
    assert hasattr(graph, "edge_index")
    assert torch.isclose(graph.x, x).all()
    if isinstance(x, LabelTensor):
        assert isinstance(graph.x, LabelTensor)
        assert graph.x.labels == x.labels
    else:
        assert isinstance(graph.pos, torch.Tensor)
    assert torch.isclose(graph.pos, pos).all()
    if isinstance(pos, LabelTensor):
        assert isinstance(graph.pos, LabelTensor)
        assert graph.pos.labels == pos.labels
    else:
        assert isinstance(graph.pos, torch.Tensor)
    if not loop:
        assert (
            len(
                torch.nonzero(
                    graph.edge_index[0] == graph.edge_index[1], as_tuple=True
                )[0]
            )
            == 0
        )  # Detect self loops


@pytest.mark.parametrize(
    "x, pos",
    [
        (torch.rand(10, 2), torch.rand(10, 3)),
        (
            LabelTensor(torch.rand(10, 2), ["u", "v"]),
            LabelTensor(torch.rand(10, 3), ["x", "y", "z"]),
        ),
    ],
)
def test_build_radius_graph_edge_attr(x, pos):
    graph = RadiusGraph(x=x, pos=pos, radius=0.5, edge_attr=True)
    assert hasattr(graph, "x")
    assert hasattr(graph, "pos")
    assert hasattr(graph, "edge_index")
    assert torch.isclose(graph.x, x).all()
    if isinstance(x, LabelTensor):
        assert isinstance(graph.x, LabelTensor)
        assert graph.x.labels == x.labels
    else:
        assert isinstance(graph.pos, torch.Tensor)
    assert torch.isclose(graph.pos, pos).all()
    if isinstance(pos, LabelTensor):
        assert isinstance(graph.pos, LabelTensor)
        assert graph.pos.labels == pos.labels
    else:
        assert isinstance(graph.pos, torch.Tensor)
    assert hasattr(graph, "edge_attr")
    assert isinstance(graph.edge_attr, torch.Tensor)
    assert graph.edge_attr.shape[-1] == 3
    assert graph.edge_attr.shape[0] == graph.edge_index.shape[1]


@pytest.mark.parametrize(
    "x, pos",
    [
        (torch.rand(10, 2), torch.rand(10, 3)),
        (
            LabelTensor(torch.rand(10, 2), ["u", "v"]),
            LabelTensor(torch.rand(10, 3), ["x", "y", "z"]),
        ),
    ],
)
def test_build_radius_graph_custom_edge_attr(x, pos):
    graph = RadiusGraph(
        x=x,
        pos=pos,
        radius=0.5,
        edge_attr=True,
        custom_edge_func=build_edge_attr,
    )
    assert hasattr(graph, "x")
    assert hasattr(graph, "pos")
    assert hasattr(graph, "edge_index")
    assert torch.isclose(graph.x, x).all()
    if isinstance(x, LabelTensor):
        assert isinstance(graph.x, LabelTensor)
        assert graph.x.labels == x.labels
    else:
        assert isinstance(graph.pos, torch.Tensor)
    assert torch.isclose(graph.pos, pos).all()
    if isinstance(pos, LabelTensor):
        assert isinstance(graph.pos, LabelTensor)
        assert graph.pos.labels == pos.labels
    else:
        assert isinstance(graph.pos, torch.Tensor)
    assert hasattr(graph, "edge_attr")
    assert isinstance(graph.edge_attr, torch.Tensor)
    assert graph.edge_attr.shape[-1] == 6
    assert graph.edge_attr.shape[0] == graph.edge_index.shape[1]


@pytest.mark.parametrize(
    "x, pos",
    [
        (torch.rand(10, 2), torch.rand(10, 3)),
        (
            LabelTensor(torch.rand(10, 2), ["u", "v"]),
            LabelTensor(torch.rand(10, 3), ["x", "y", "z"]),
        ),
    ],
)
@pytest.mark.parametrize("loop", [True, False])
def test_build_knn_graph(x, pos, loop):
    graph = KNNGraph(x=x, pos=pos, neighbours=2, loop=loop)
    assert hasattr(graph, "x")
    assert hasattr(graph, "pos")
    assert hasattr(graph, "edge_index")
    assert torch.isclose(graph.x, x).all()
    if isinstance(x, LabelTensor):
        assert isinstance(graph.x, LabelTensor)
        assert graph.x.labels == x.labels
    else:
        assert isinstance(graph.pos, torch.Tensor)
    assert torch.isclose(graph.pos, pos).all()
    if isinstance(pos, LabelTensor):
        assert isinstance(graph.pos, LabelTensor)
        assert graph.pos.labels == pos.labels
    else:
        assert isinstance(graph.pos, torch.Tensor)
    assert graph.edge_attr is None
    self_loops = len(
        torch.nonzero(
            graph.edge_index[0] == graph.edge_index[1], as_tuple=True
        )[0]
    )
    if loop:
        assert self_loops != 0
    else:
        assert self_loops == 0


@pytest.mark.parametrize(
    "x, pos",
    [
        (torch.rand(10, 2), torch.rand(10, 3)),
        (
            LabelTensor(torch.rand(10, 2), ["u", "v"]),
            LabelTensor(torch.rand(10, 3), ["x", "y", "z"]),
        ),
    ],
)
def test_build_knn_graph_edge_attr(x, pos):
    graph = KNNGraph(x=x, pos=pos, neighbours=2, edge_attr=True)
    assert hasattr(graph, "x")
    assert hasattr(graph, "pos")
    assert hasattr(graph, "edge_index")
    assert torch.isclose(graph.x, x).all()
    if isinstance(x, LabelTensor):
        assert isinstance(graph.x, LabelTensor)
        assert graph.x.labels == x.labels
    else:
        assert isinstance(graph.pos, torch.Tensor)
    assert torch.isclose(graph.pos, pos).all()
    if isinstance(pos, LabelTensor):
        assert isinstance(graph.pos, LabelTensor)
        assert graph.pos.labels == pos.labels
    else:
        assert isinstance(graph.pos, torch.Tensor)
    assert isinstance(graph.edge_attr, torch.Tensor)
    assert graph.edge_attr.shape[-1] == 3
    assert graph.edge_attr.shape[0] == graph.edge_index.shape[1]


@pytest.mark.parametrize(
    "x, pos",
    [
        (torch.rand(10, 2), torch.rand(10, 3)),
        (
            LabelTensor(torch.rand(10, 2), ["u", "v"]),
            LabelTensor(torch.rand(10, 3), ["x", "y", "z"]),
        ),
    ],
)
def test_build_knn_graph_custom_edge_attr(x, pos):
    graph = KNNGraph(
        x=x,
        pos=pos,
        neighbours=2,
        edge_attr=True,
        custom_edge_func=build_edge_attr,
    )
    assert hasattr(graph, "x")
    assert hasattr(graph, "pos")
    assert hasattr(graph, "edge_index")
    assert torch.isclose(graph.x, x).all()
    if isinstance(x, LabelTensor):
        assert isinstance(graph.x, LabelTensor)
        assert graph.x.labels == x.labels
    else:
        assert isinstance(graph.pos, torch.Tensor)
    assert torch.isclose(graph.pos, pos).all()
    if isinstance(pos, LabelTensor):
        assert isinstance(graph.pos, LabelTensor)
        assert graph.pos.labels == pos.labels
    else:
        assert isinstance(graph.pos, torch.Tensor)
    assert isinstance(graph.edge_attr, torch.Tensor)
    assert graph.edge_attr.shape[-1] == 6
    assert graph.edge_attr.shape[0] == graph.edge_index.shape[1]


@pytest.mark.parametrize(
    "x, pos, y",
    [
        (torch.rand(10, 2), torch.rand(10, 3), torch.rand(10, 4)),
        (
            LabelTensor(torch.rand(10, 2), ["u", "v"]),
            LabelTensor(torch.rand(10, 3), ["x", "y", "z"]),
            LabelTensor(torch.rand(10, 4), ["a", "b", "c", "d"]),
        ),
    ],
)
def test_additional_params(x, pos, y):
    edge_index = torch.tensor(
        [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [1, 2, 3, 4, 5, 6, 7, 8, 9, 0]],
        dtype=torch.int64,
    )
    graph = Graph(x=x, pos=pos, edge_index=edge_index, y=y)
    assert hasattr(graph, "y")
    assert torch.isclose(graph.y, y).all()
    if isinstance(y, LabelTensor):
        assert isinstance(graph.y, LabelTensor)
        assert graph.y.labels == y.labels
    else:
        assert isinstance(graph.y, torch.Tensor)
    assert torch.isclose(graph.y, y).all()
    if isinstance(y, LabelTensor):
        assert isinstance(graph.y, LabelTensor)
        assert graph.y.labels == y.labels
    else:
        assert isinstance(graph.y, torch.Tensor)


@pytest.mark.parametrize(
    "x, pos, y",
    [
        (torch.rand(10, 2), torch.rand(10, 3), torch.rand(10, 4)),
        (
            LabelTensor(torch.rand(10, 2), ["u", "v"]),
            LabelTensor(torch.rand(10, 3), ["x", "y", "z"]),
            LabelTensor(torch.rand(10, 4), ["a", "b", "c", "d"]),
        ),
    ],
)
def test_additional_params_radius_graph(x, pos, y):
    graph = RadiusGraph(x=x, pos=pos, radius=0.5, y=y)
    assert hasattr(graph, "y")
    assert torch.isclose(graph.y, y).all()
    if isinstance(y, LabelTensor):
        assert isinstance(graph.y, LabelTensor)
        assert graph.y.labels == y.labels
    else:
        assert isinstance(graph.y, torch.Tensor)
    assert torch.isclose(graph.y, y).all()
    if isinstance(y, LabelTensor):
        assert isinstance(graph.y, LabelTensor)
        assert graph.y.labels == y.labels
    else:
        assert isinstance(graph.y, torch.Tensor)


@pytest.mark.parametrize(
    "x, pos, y",
    [
        (torch.rand(10, 2), torch.rand(10, 3), torch.rand(10, 4)),
        (
            LabelTensor(torch.rand(10, 2), ["u", "v"]),
            LabelTensor(torch.rand(10, 3), ["x", "y", "z"]),
            LabelTensor(torch.rand(10, 4), ["a", "b", "c", "d"]),
        ),
    ],
)
def test_additional_params_knn_graph(x, pos, y):
    graph = KNNGraph(x=x, pos=pos, neighbours=3, y=y)
    assert hasattr(graph, "y")
    assert torch.isclose(graph.y, y).all()
    if isinstance(y, LabelTensor):
        assert isinstance(graph.y, LabelTensor)
        assert graph.y.labels == y.labels
    else:
        assert isinstance(graph.y, torch.Tensor)
    assert torch.isclose(graph.y, y).all()
    if isinstance(y, LabelTensor):
        assert isinstance(graph.y, LabelTensor)
        assert graph.y.labels == y.labels
    else:
        assert isinstance(graph.y, torch.Tensor)



"""Module for defining various general equations."""

from .equation import Equation
from ..operator import grad, div, laplacian


class FixedValue(Equation):
    """
    Equation to enforce a fixed value. Can be used to enforce Dirichlet Boundary
    conditions.
    """

    def __init__(self, value, components=None):
        """
        Initialization of the :class:`FixedValue` class.

        :param float value: The fixed value to be enforced.
        :param list[str] components: The name of the output variables for which
            the fixed value condition is applied. It should be a subset of the
            output labels. If ``None``, all output variables are considered.
            Default is ``None``.
        """

        def equation(input_, output_):
            """
            Definition of the equation to enforce a fixed value.

            :param LabelTensor input_: Input points where the equation is
                evaluated.
            :param LabelTensor output_: Output tensor, eventually produced by a
                :class:`torch.nn.Module` instance.
            :return: The computed residual of the equation.
            :rtype: LabelTensor
            """
            if components is None:
                return output_ - value
            return output_.extract(components) - value

        super().__init__(equation)


class FixedGradient(Equation):
    """
    Equation to enforce a fixed gradient for a specific condition.
    """

    def __init__(self, value, components=None, d=None):
        """
        Initialization of the :class:`FixedGradient` class.

        :param float value: The fixed value to be enforced to the gradient.
        :param list[str] components: The name of the output variables for which
            the fixed gradient condition is applied. It should be a subset of
            the output labels. If ``None``, all output variables are considered.
            Default is ``None``.
        :param list[str] d: The name of the input variables on which the
            gradient is computed. It should be a subset of the input labels.
            If ``None``, all the input variables are considered.
            Default is ``None``.
        """

        def equation(input_, output_):
            """
            Definition of the equation to enforce a fixed gradient.

            :param LabelTensor input_: Input points where the equation is
                evaluated.
            :param LabelTensor output_: Output tensor, eventually produced by a
                :class:`torch.nn.Module` instance.
            :return: The computed residual of the equation.
            :rtype: LabelTensor
            """
            return grad(output_, input_, components=components, d=d) - value

        super().__init__(equation)


class FixedFlux(Equation):
    """
    Equation to enforce a fixed flux, or divergence, for a specific condition.
    """

    def __init__(self, value, components=None, d=None):
        """
        Initialization of the :class:`FixedFlux` class.

        :param float value: The fixed value to be enforced to the flux.
        :param list[str] components: The name of the output variables for which
            the fixed flux condition is applied. It should be a subset of the
            output labels. If ``None``, all output variables are considered.
            Default is ``None``.
        :param list[str] d: The name of the input variables on which the flux
            is computed. It should be a subset of the input labels. If ``None``,
            all the input variables are considered. Default is ``None``.
        """

        def equation(input_, output_):
            """
            Definition of the equation to enforce a fixed flux.

            :param LabelTensor input_: Input points where the equation is
                evaluated.
            :param LabelTensor output_: Output tensor, eventually produced by a
                :class:`torch.nn.Module` instance.
            :return: The computed residual of the equation.
            :rtype: LabelTensor
            """
            return div(output_, input_, components=components, d=d) - value

        super().__init__(equation)


class Laplace(Equation):
    """
    Equation to enforce a null laplacian for a specific condition.
    """

    def __init__(self, components=None, d=None):
        """
        Initialization of the :class:`Laplace` class.

        :param list[str] components: The name of the output variables for which
            the null laplace condition is applied. It should be a subset of the
            output labels. If ``None``, all output variables are considered.
            Default is ``None``.
        :param list[str] d: The name of the input variables on which the
            laplacian is computed. It should be a subset of the input labels.
            If ``None``, all the input variables are considered.
            Default is ``None``.
        """

        def equation(input_, output_):
            """
            Definition of the equation to enforce a null laplacian.

            :param LabelTensor input_: Input points where the equation is
                evaluated.
            :param LabelTensor output_: Output tensor, eventually produced by a
                :class:`torch.nn.Module` instance.
            :return: The computed residual of the equation.
            :rtype: LabelTensor
            """
            return laplacian(output_, input_, components=components, d=d)

        super().__init__(equation)



"""Module for the Multi Feed Forward model class."""

from abc import ABC, abstractmethod
import torch
from .feed_forward import FeedForward


class MultiFeedForward(torch.nn.Module, ABC):
    """
    Multi Feed Forward neural network model class.

    This model allows to create a network with multiple Feed Forward neural
    networks combined together. The user is required to define the ``forward``
    method to choose how to combine the networks.
    """

    def __init__(self, ffn_dict):
        """
        Initialization of the :class:`MultiFeedForward` class.

        :param dict ffn_dict: A dictionary containing the Feed Forward neural
            networks to be combined.
        :raises TypeError: If the input is not a dictionary.
        """
        super().__init__()

        if not isinstance(ffn_dict, dict):
            raise TypeError

        for name, constructor_args in ffn_dict.items():
            setattr(self, name, FeedForward(**constructor_args))

    @abstractmethod
    def forward(self, *args, **kwargs):
        """
        Forward pass for the :class:`MultiFeedForward` model.

        The user is required to define this method to choose how to combine the
        networks.
        """
